/**
Action Script library. Math class implementation
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_math.hh"
#include <math.h>

typedef void (ASMath::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASMath::asabs,			// 0
	&ASMath::asacos,		// 1
	&ASMath::asasin,		// 2
	&ASMath::asatan,		// 3
	&ASMath::asatan2,		// 4
	&ASMath::asceil,		// 5
	&ASMath::ascos,			// 6
	&ASMath::asexp,			// 7
	&ASMath::asfloor,		// 8
	&ASMath::aslog,			// 9
	&ASMath::asmax,			// 10
	&ASMath::asmin,			// 11
	&ASMath::aspow,			// 12
	&ASMath::random,		// 13
	&ASMath::asround,		// 14
	&ASMath::assin,			// 15
	&ASMath::assqrt,		// 16
	&ASMath::astan,			// 17
	&ASMath::init,			// 18
	// Need to be at the end.
	&ASMath::addProperty,		
	&ASMath::registerClass,	
	&ASMath::AStoString,		
	&ASMath::unwatch,			
	&ASMath::valueOf,			
	&ASMath::watch			
};

/**
 * The constructor
 **/
ASMath::ASMath(Script *s, Graphics *g) : ASObject(s,g,L"Math") {
	setMemberNoClone(L"E",new DoubleObject(M_E));
	setMemberNoClone(L"LN2",new DoubleObject(M_LN2));
	setMemberNoClone(L"LOG2E",new DoubleObject(M_LOG2E));
	setMemberNoClone(L"LN10",new DoubleObject(M_LN10));
	setMemberNoClone(L"LOG10E",new DoubleObject(M_LOG10E));
	setMemberNoClone(L"PI",new DoubleObject(M_PI));
	setMemberNoClone(L"SQRT1_2",new DoubleObject(M_SQRT1_2));
	setMemberNoClone(L"SQRT2",new DoubleObject(M_SQRT2));
}

/**
 * The destructor
 **/
ASMath::~ASMath() {
}

/**
 * Return a function descriptor
 */
Function *ASMath::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"abs") 	{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"acos") 	{func->setFunctionIndex(1); func->setNumParams(1); }
	if (memberName == L"asin") 	{func->setFunctionIndex(2); func->setNumParams(1);}
	if (memberName == L"atan") 	{func->setFunctionIndex(3); func->setNumParams(1);}
	if (memberName == L"atan2") {func->setFunctionIndex(4); func->setNumParams(2);}
	if (memberName == L"ceil") 	{func->setFunctionIndex(5); func->setNumParams(1);}
	if (memberName == L"cos") 	{func->setFunctionIndex(6); func->setNumParams(1);}
	if (memberName == L"exp") 	{func->setFunctionIndex(7); func->setNumParams(1);}
	if (memberName == L"floor") {func->setFunctionIndex(8); func->setNumParams(1);}
	if (memberName == L"log") 	{func->setFunctionIndex(9); func->setNumParams(1);}
	if (memberName == L"max") 	{func->setFunctionIndex(10); func->setNumParams(2);}
	if (memberName == L"min") 	{func->setFunctionIndex(11); func->setNumParams(2);}
	if (memberName == L"pow") 	{func->setFunctionIndex(12); func->setNumParams(2);}
	if (memberName == L"random"){func->setFunctionIndex(13); func->setNumParams(0);}
	if (memberName == L"round") {func->setFunctionIndex(14); func->setNumParams(1);}
	if (memberName == L"sin") 	{func->setFunctionIndex(15); func->setNumParams(1);}
	if (memberName == L"sqrt") 	{func->setFunctionIndex(16); func->setNumParams(1);}
	if (memberName == L"tan") 	{func->setFunctionIndex(17); func->setNumParams(1);}
	if (memberName == L"Math") 	{func->setFunctionIndex(18); func->setNumParams(0);}
	finishFunctionSetup(memberName,19,func);
	return func;
}
 
/**
 * Call class built in functions
 * exotic c++ call, but fast
 **/
void ASMath::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASMath::init() {
	puts("math constructor");
}

/**
 * computes and returns an absolute value for the number specified by the parameter x.
 **/
void ASMath::asabs() {
	ValueObject *p = script->getRegister(1);
	script->push(new FloatObject(fabs(p->toNumeric())));
}

/**
 * Return ACOS
 **/
void ASMath::asacos() {
	double a = script->getRegister(1)->toNumeric();
	if (fabs(a) <= 1.0)
		script->push(new DoubleObject(acos(a)));
	else
		script->push(new StringObject(L"NaN"));
}

/**
 * Return ASIN
 **/
void ASMath::asasin() {
	double a = script->getRegister(1)->toNumeric();
	if (fabs(a) <= 1.0)
		script->push(new DoubleObject(asin(a)));
	else
		script->push(new StringObject(L"NaN"));
}

/**
 * Return ATAN
 **/
void ASMath::asatan() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(atan(a)));
}

/**
 * Return ATAN2
 **/
void ASMath::asatan2() {
	double x = script->getRegister(1)->toNumeric();
	double y = script->getRegister(2)->toNumeric();
	script->push(new DoubleObject(atan2(x,y)));
}

/**
 * Rounds a number up to the nearest integer.
 **/
void ASMath::asceil() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new FloatObject(ceil(a)));
}

/**
 * Return cosine(radian)
 **/
void ASMath::ascos() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(cos(a)));	
}

/**
 * Return EXP
 **/
void ASMath::asexp() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(exp(a)));	
}

/**
 *  Rounds a number down to the nearest integer.
 **/
void ASMath::asfloor() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(floor(a)));	
}

/**
 * Return LOG
 **/
void ASMath::aslog() {
	double a = script->getRegister(1)->toNumeric();
	if (a > 0) script->push(new DoubleObject(log(a)));
	else script->push(new StringObject(L"NaN"));
}

/**
 * MAX of two numbers
 **/
void ASMath::asmax() {
	double a = script->getRegister(1)->toNumeric();
	double b = script->getRegister(2)->toNumeric();
	script->push(new DoubleObject(a>b ? a : b));
}

/**
 * MIN of two numbers 
 **/
void ASMath::asmin() {
	double a = script->getRegister(1)->toNumeric();
	double b = script->getRegister(2)->toNumeric();
	script->push(new DoubleObject(a < b ? a : b));
}

/**
 * Return POW
 **/
void ASMath::aspow() {
	double x = script->getRegister(1)->toNumeric();
	double y = script->getRegister(2)->toNumeric();
	if (x > 0)	script->push(new DoubleObject(pow(x,y)));
	else script->push(new StringObject(L"NaN"));	// How to handle complex numbers ?
}

/**
 * Method; returns n, where 0 <= n < 1.
 **/
void ASMath::random() {
	script->push(new DoubleObject(drand48()));
}

/**
 * Rounds to the nearest integer.
 **/
void ASMath::asround() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(round(a)));	
}

/**
 *  Return sinus(radian)
 **/
void ASMath::assin() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(sin(a)));	
}

/**
 * Return SQRT
 **/
void ASMath::assqrt() {
	double x = script->getRegister(1)->toNumeric();
	if (x >= 0)	script->push(new DoubleObject(sqrt(x)));
	else script->push(new StringObject(L"NaN"));	
}

/**
 *  Return ATAN
 **/
void ASMath::astan() {
	double a = script->getRegister(1)->toNumeric();
	script->push(new DoubleObject(atan(a)));	
}
