/**
Action Script library. Sound class.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_sound.hh"
#include <vector>

// This function is the start of separate streaming sound threads
extern "C" void *ProcessStreamingSound(void*);
void fillAudioHandler(void *udata, Uint8 *stream, int len); // SDL C audio handler
static pthread_mutex_t queueSynchro = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;	// Used in the audioQueues wrapper class.
static pthread_mutex_t decodingLock = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;	// To synchronize access to decodingFinished
static pthread_cond_t emptyStream = PTHREAD_COND_INITIALIZER;
static pthread_cond_t fullStream = PTHREAD_COND_INITIALIZER;
static pthread_cond_t putWaiting = PTHREAD_COND_INITIALIZER;		 // Used to signal ~AudioQueues that AudioQueues::push() is finished


typedef void (ASSound::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASSound::init,				// 0
	&ASSound::attachSound,		// 1
	&ASSound::getBytesLoaded,	// 2
	&ASSound::getBytesTotal,	// 3
	&ASSound::getPan,			// 4
	&ASSound::getTransform,		// 5
	&ASSound::getVolume,		// 6
	&ASSound::loadSound,		// 7
	&ASSound::setPan,			// 8
	&ASSound::setTransform,		// 9
	&ASSound::setVolume,		// 10
	&ASSound::start,			// 11
	&ASSound::stop,				// 12
	&ASSound::onID3,			// 13
	&ASSound::onLoad,			// 14
	&ASSound::onSoundComplete,	// 15
	// Need to be at the end.
	&ASSound::addProperty,		
	&ASSound::registerClass,	
	&ASSound::AStoString,		
	&ASSound::unwatch,			
	&ASSound::valueOf,			
	&ASSound::watch
};

#define SOUND_BUFFER_SIZE 32768

/**
 * The constructor
 **/
ASSound::ASSound(Script *s, Graphics *g) 
	: ASObject(s,g,L"Sound"),
	soundError(false),
	soundOpened(false),
	soundInProgress(false),
	soundInitialized(false)
{
	// Check little/big endian for sound buffer filling.
	char tmp[2]; tmp[0] = 1 ; tmp[1] = 2; short *c = (short*)tmp;
	timer = s->getTimer();
	isLittleEndian = *c == 0x12 ? false: true;
	// Used to check/set whether streamingSound is currently playing
	audioQueues = new AudioQueues(timer);
	decodingFinished = new map<UI16, bool>;
	// Set attributes for the audio thread
	if(pthread_attr_init(&pthread_attr)){
		wcout << "problems creating thread attributes\n";
		exit(1);
	}
	if(pthread_attr_setdetachstate(&pthread_attr, PTHREAD_CREATE_DETACHED)){
		wcout << "problems creating thread attributes\n";
		exit(1);
	}
	if(pthread_attr_setschedpolicy(&pthread_attr, SCHED_RR)){
		wcout << "problems creating thread attributes\n";
		exit(1);
	}
}

/**
 * The destructor. 
 *
 * We can't destroy the sound manager if there is a sound buffer fill interrupt.
 * It will wait for the sound buffers to be filled.
 **/
ASSound::~ASSound() {
	stop();
	for (IntSOUNDTRACKHashMap::iterator it=soundTracks.begin(); it!= soundTracks.end(); it++) {
		SOUNDTRACK *st = (SOUNDTRACK*)(*it).second;
		delete st;
	}
	soundTracks.clear();
	soundTags.clear();
	pthread_attr_destroy(&pthread_attr);
}

/**
 * Return a function descriptor
 */
Function *ASSound::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Sound") 		{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"attachSound") 	{func->setFunctionIndex(1); func->setNumParams(1); }
	if (memberName == L"getBytesLoaded"){func->setFunctionIndex(2); func->setNumParams(0); }
	if (memberName == L"getBytesTotal") {func->setFunctionIndex(3); func->setNumParams(0); }
	if (memberName == L"getPan") 		{func->setFunctionIndex(4); func->setNumParams(0); }
	if (memberName == L"getTransform") 	{func->setFunctionIndex(5); func->setNumParams(0); }
	if (memberName == L"getVolume") 	{func->setFunctionIndex(6); func->setNumParams(0); }
	if (memberName == L"loadSound") 	{func->setFunctionIndex(7); func->setNumParams(2); }
	if (memberName == L"setPan") 		{func->setFunctionIndex(8); func->setNumParams(1); }
	if (memberName == L"setTransform") 	{func->setFunctionIndex(9); func->setNumParams(1); }
	if (memberName == L"setVolume") 	{func->setFunctionIndex(10); func->setNumParams(1); }
	if (memberName == L"start") 		{func->setFunctionIndex(11); func->setNumParams(2); }
	if (memberName == L"stop") 			{func->setFunctionIndex(12); func->setNumParams(1); }
	if (memberName == L"onID3") 		{func->setFunctionIndex(13); func->setNumParams(1); }
	if (memberName == L"onLoad") 		{func->setFunctionIndex(14); func->setNumParams(1); }
	if (memberName == L"onSoundComplete"){func->setFunctionIndex(15); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,16,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASSound::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}
/**
 *
 **/
void ASSound::init() {
	puts("TODO - ASSound::init()");
}

/**
 * Initialize a sound stream head given a SoundStreamHead Tag
 **/
void ASSound::init(SoundStreamHeadTag*t) {
	if(soundError)
		return;
	if (!soundInitialized)  {
		if( SDL_Init(SDL_INIT_AUDIO) <0 ) {
		  printf("Unable to init SDL: %s\n", SDL_GetError());
		  soundError = true;
		  return;
		}
		soundInitialized = true;
	}

	// Register Audio codecs
	av_register_all();

	// Get required audio output configuration, and configure the SDL audiospec.
	playbackSoundRate = t->getPlaybackSoundRate();
	streamSoundCompression = t->getStreamSoundCompression();
	streamSoundRate = t->getStreamSoundRate();
	streamSoundSampleCount = t->getStreamSoundSampleCount();
	latencySeek = t->getLatencySeek();
	playbackSoundSize = t->getPlaybackSoundSize();
	playbackSoundType = t->getPlaybackSoundType();
	streamSoundSize = t->getStreamSoundSize();
	streamSoundType = t->getStreamSoundType();

	switch (streamSoundRate) {
		case 0: audioSpec.freq = 5512;  break;
		case 1: audioSpec.freq = 11025; break;
		case 2: audioSpec.freq = 22050; break;
		case 3: audioSpec.freq = 44100; break;
		default: audioSpec.freq = 22050;
	}
	audioSpec.format = AUDIO_S16SYS;
	audioSpec.channels = streamSoundType ? 2 : 1;
	audioSpec.callback = fillAudioHandler;
	audioSpec.userdata = this;
	audioSpec.samples = 1024;
	
	// Audio specs should control timeline only if it is from main timeline
	if(t->getCurSpriteId() == 0)
		timer->setSoundParams(audioSpec.samples, audioSpec.freq, streamSoundSampleCount);
}

/**
 * Attaches the sound specified in the parameter.
 **/
void ASSound::attachSound() {
	puts("TODO - ASSound::attachSound()");
	noReturnValue();
}

/**
 *  Returns the number of bytes loaded for the specified sound.
 **/
void ASSound::getBytesLoaded() {
	puts("TODO - ASSound::getBytesLoaded()");
	noReturnValue();
}

/**
 * Returns the size of the sound in bytes.
 **/
void ASSound::getBytesTotal() {
	puts("TODO - ASSound::getBytesTotal()");
	noReturnValue();
}

/**
 * Returns the value of the previous setPan() call.
 **/
void ASSound::getPan() {
	puts("TODO - ASSound::getPan()");
	noReturnValue();
}

/**
 * Returns the value of the previous setTransform() call.
 **/
void ASSound::getTransform() {
	puts("TODO - ASSound::getTransform()");
	noReturnValue();
}

/**
 * Returns the value of the previous setVolume() call.
 **/
void ASSound::getVolume() {
	puts("TODO - ASSound::getVolume()");
	noReturnValue();
}

/**
 * Loads an MP3 file into Flash Player.
 **/
void ASSound::loadSound() {
	puts("TODO - ASSound::loadSound()");
	noReturnValue();
}

/**
 * Sets the left/right balance of the sound.
 **/
void ASSound::setPan() {
	puts("TODO - ASSound::setPan()");
	noReturnValue();
}

/**
 * Sets the amount of each channel, left and right, to be played in each speaker.
 **/
void ASSound::setTransform() {
	puts("TODO - ASSound::setTransform()");
	noReturnValue();
}

/**
 * Sets the volume level for a sound.
 **/
void ASSound::setVolume() {
	puts("TODO - ASSound::setVolume()");
	noReturnValue();
}

/**
 * Starts playing a sound from the beginning or, optionally, from an offset point set in the parameter.
 **/
void ASSound::start() {
	puts("TODO - ASSound::start()");
	noReturnValue();
}


/**
 * Stops all sounds currently playing.
 **/
void ASSound::stop() {
	if(soundError)
		return;

	pthread_mutex_lock(&audioSynchro);
	setSoundInProgress(false);
	pthread_mutex_unlock(&audioSynchro);

	// All streaming sounds are now told to exit
	vector<UI16> *sprites = audioQueues->getSpriteIds();
	typedef vector<UI16>::iterator iter;
	for(iter strIt = sprites->begin(); strIt != sprites->end(); strIt++){
		if(timer->getSoundStreamPlaying(*strIt) == PLAYING);
			timer->setSoundStreamPlaying(*strIt, EXITING);
	}
	audioQueues->signal();
	timer->stop();

	// Waiting till they have finished
	for(iter strIt = sprites->begin(); strIt != sprites->end(); strIt++){
		timer->waitForStreamEnd(*strIt);
	}
	delete sprites;
	
	if (soundOpened) {
		SDL_CloseAudio();
		soundOpened = false;
	}
	noReturnValue();
}

/**
 * Create a new sound track given a DefineSound tag
 **/
void ASSound::registerSoundTrack(DefineSoundTag *t) {
	soundTags[t->getSoundId()] = t;
}

/**
 * Call the decoding libraries and put decoded sound in a buffer list.
 *
 * hasDecoder is true when called from processSoundStreamBloc.
 * Means pCodec and pCodecCtx should not be touched
 **/
int ASSound::decodeSoundTrack(SOUNDTRACK *s, bool hasDecoder) {
	int requestedSize = 0;
	s->ptrData = s->soundData;
	SoundBlock *sb = new SoundBlock();
	int outSize = 0;
	int len = 0;
	uint8_t *outbuf=NULL;
	switch (s->soundFormat) {
		case 0: {		// uncompressed
			requestedSize = s->soundDataSize;
			outbuf = (uint8_t*)malloc(requestedSize);
			if (isLittleEndian) {
				for (int i=0; i < requestedSize; i+= 2) {
					outbuf[i]   = s->ptrData[i+1];
					outbuf[i+1] = s->ptrData[i];
				}
			} else {
				for (int i=0; i < requestedSize; i+= 2) {
					outbuf[i]   = s->ptrData[i];
					outbuf[i+1] = s->ptrData[i+1];
				}
			}
			outSize = len = requestedSize;
			puts("SOUND type uncompressed");
			break;
		}
		case 1: {		// ADPCM
			s->adpcmDecoder = new Adpcm(s->ptrData,s->soundType);
			requestedSize = s->soundDataSize;
			outbuf = (uint8_t*)malloc(requestedSize*sizeof(short)*(s->soundType==0?1:2));
			len = s->adpcmDecoder->decompress((short *)outbuf,requestedSize);
			outSize = len;
			delete s->adpcmDecoder;
			//puts("SOUND type ADPCM");
			break;
		}
		case 2: { 		// MP3
			// It seems ffmpeg cannot handle allocating and deallocating decoders fast and repeatedly
			// Therefore, this is only done for eventSounds
			// When called from processSoundStreamBlock, the decoder is assumed existing (hasDecoder is true)
                        // The decoder context is later used in decodeNextBlock, and should not be deallocated
			if(!hasDecoder){
 	  			s->pCodec=avcodec_find_decoder(CODEC_ID_MP3);
				if (!s->pCodec) {
					fprintf(stderr, "MP3 codec not found in ffmpeg\n");
					return -1;
				}
				s->pCodecCtx= avcodec_alloc_context();
		    		if (avcodec_open(s->pCodecCtx, s->pCodec) <0) { puts(" Could not open Codec"); return -1; }
			}

			requestedSize = s->soundDataSize > 4096 ? 4096 : s->soundDataSize;
			outbuf = (uint8_t*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);

			len = avcodec_decode_audio(s->pCodecCtx, (short *)outbuf, &outSize, s->ptrData, requestedSize);

			// Errors from bugs in swf-file or this program. Hopefully just former :)
			if(outSize < 0) wcout << "mp3 decoding error: outSize = " << outSize << "\n";
			if (len < 0) {
				fprintf(stderr, "Error while decoding SOUND\n");
				return -1;
			}

			//puts("SOUND type MP3");
			break;
		}
		case 3:	{		//  uncompressed little-endian
			requestedSize = s->soundDataSize;
			outbuf = (uint8_t*)malloc(requestedSize);
			if (isLittleEndian) {
				for (int i=0; i < requestedSize; i+= 2) {
					outbuf[i]   = s->ptrData[i];
					outbuf[i+1] = s->ptrData[i+1];
				}
			} else {
				for (int i=0; i < requestedSize; i+= 2) {
					outbuf[i]   = s->ptrData[i+1];
					outbuf[i+1] = s->ptrData[i];
				}
			}
			outSize = len = requestedSize;
			//puts("SOUND type uncompressed little-endian");
			break;
		}
		case 6: {		//  Nellymoser
			puts("SOUND type Nellymoser not supported yet");
			return -1;
			break;
		}
	}

	// Update the sound buffer list.
	sb->buffer = outbuf;
	sb->dpos = 0;
	sb->dlen = outSize;
	s->outputBuffers.push_back(sb);
	s->ptrData += len;
	return len;
}


/**
 * Start playing a sound track
 **/
void ASSound::startSound(int soundId,SOUNDINFO *p) {
	if(soundError)
		return;
	DefineSoundTag *t = soundTags[soundId];
	SOUNDTRACK *s = new SOUNDTRACK;
	s->soundFormat = t->getSoundFormat();
	s->soundRate= t->getSoundRate();
	s->soundSize= t->getSoundSize();
	s->soundType= t->getSoundType();
	s->soundSampleCount= t->getSoundSampleCount();
	s->soundData= t->getSoundData();
	s->soundDataSize = t->getSoundDataSize();
	s->isSoundStreamBlock = false;
	decodeSoundTrack(s, false);
	soundTracks[t->getSoundId()] = s;

	if (!soundOpened) {
		switch (s->soundRate) {
			case 0: audioSpec.freq = 5512;  break;
			case 1: audioSpec.freq = 11025; break;
			case 2: audioSpec.freq = 22050; break;
			case 3: audioSpec.freq = 44100; break;
			default: audioSpec.freq = 22050;
		}
		audioSpec.format = AUDIO_S16SYS;
    		audioSpec.channels = s->soundType ? 2 : 1;
    		audioSpec.callback = fillAudioHandler;
    		audioSpec.userdata = this;
    		audioSpec.samples = 1024;

		if (!openSound())
			return;
	}
	soundInProgress = true;
	SDL_PauseAudio(0);
}

/**
 * Decode next block of data
 *
 * This call back is only used by MP3 as that's the only format that allow that.
 **/
void ASSound::decodeNextBlock(SOUNDTRACK *s) {
		int remainingSize = (s->soundData + s->soundDataSize) - s->ptrData;
		if (remainingSize <= 0 || s->soundFormat != 2) return;

		int outSize;
		uint8_t *outbuf = (uint8_t*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
		int len = avcodec_decode_audio(s->pCodecCtx, (short *)outbuf, &outSize, s->ptrData, remainingSize);
		if (len < 0) {
			fprintf(stderr, "Error while decoding SOUND\n");
			free(outbuf);
			return;
		}
		if (outSize > 0) { //if a frame has been decoded, output it
			SoundBlock *sb = new SoundBlock();
			sb->buffer = (uint8_t*)realloc(outbuf,outSize+1);	
			sb->dpos = 0;
			sb->dlen = outSize;
			s->outputBuffers.push_back(sb);
		}
		s->ptrData += len;
}

bool ASSound::openSound(){
	if(soundError)
		return false;
	if (!soundOpened) {
		if ( SDL_OpenAudio(&audioSpec, NULL) < 0 ) {
			soundError = true;
			fprintf(stderr, "Unable to START SOUND: %s\n", SDL_GetError());
			return false;
		}
		soundOpened = true;
	}
	return true;
}

/**
 * Called from 'case stagSoundStreamBlock' in SwfEngine to start audio thread
*/
void ASSound::startSoundStream(SoundStreamBlockTag* stag){
	if(soundError)
		return;
	UI16 spriteId = stag->getCurSpriteId();
	streamstate State = timer->getSoundStreamPlaying(spriteId);
	if(State == PLAYING)
		return;
	PRARGS *args = new PRARGS;
	args->tag = stag;
	args->timer = timer;
	args->sound = (void*) this;
	pthread_t *trid = new pthread_t;
	
	// If there already was a thread with this sprite - kill it and start a new
	// This should not happen. Streams should be stopped before. Just extra caution
	if(State != FINISHED){
		wcout << "ASSound::startSoundStream: Error: Sound stream already playing in sprite" << spriteId << endl;
		timer->waitForStreamEnd(spriteId);
	}

	timer->setSoundStreamPlaying(spriteId, PLAYING);
	timer->setStreamIdStarted(stag->getStreamId());

	
	if (!openSound()){
		timer->setSoundStreamPlaying(spriteId, FINISHED);
		delete trid;
		wcout << "Aborting stream sound startup\n";
		return;
	}

	// Try 5 times to create the thread. Sometimes this works
	int counter = 5;

	int err = pthread_create(trid, NULL, &ProcessStreamingSound, args);
	while(err == EAGAIN && counter > 0){
		wcout << "ASSound::startSoundStream: Retrying pthread_create\n";
		err = pthread_create(trid, NULL, &ProcessStreamingSound, args);
		counter--;
	}
	if(err){
		wcout << L"Failed creating thread!. Error: " << err << endl;
		exit(1);
	}

	timer->waitForAudioFrame(spriteId, stag->getFrameNumber());
	delete trid;
	return;
}

/**
 * This is called as a separate thread from decodeSoundStreamBlock via startSoundStream
 * Sends te first SoundStreamBlock to ASSound which starts the timer
*/
extern "C" void *ProcessStreamingSound(void* pointer){

	PRARGS* args = (PRARGS*) pointer;
	SwfTimer *timer = (SwfTimer*) args->timer;
	UI16 curSpriteId = ((SoundStreamBlockTag*)args->tag)->getCurSpriteId();
	((ASSound*) args->sound)->processSoundStreamBloc((SoundStreamBlockTag*) args->tag);
	delete args;
	return NULL;
}

/**
 *Threadsafe Wrapper-functions for decodingFinished vector
 * Set by processStream when all tags have been sent to ASSound
*/
void ASSound::setDecodingFinished(UI16 id, bool finish){
	int err = pthread_mutex_lock(&decodingLock);
	if(err != 0){
			wcout << "error #" << err << " locking  decodingLock\n";
			exit(1);
	}
	(*decodingFinished)[id] = finish;
	pthread_mutex_unlock(&decodingLock);
}

/**
 * Tested by fillAudioHandler, will in case exit and signal processStream to exit
*/
bool ASSound::getDecodingFinished(UI16 id){
	int err = pthread_mutex_lock(&decodingLock);
	if(err != 0){
			wcout << "error #" << err << " locking  decodingLock\n";
			exit(1);
	}
	
	bool fin = (*decodingFinished)[id];
	pthread_mutex_unlock(&decodingLock);
	return fin;
}

void ASSound::onID3() { puts("TODO - ASSound::onID3()"); }
void ASSound::onLoad(){ puts("TODO - ASSound::onLoad()"); }
void ASSound::onSoundComplete(){ puts("TODO - ASSound::onSoundComplete()"); }


/**
 * The SoundStreamBlock tag defines sound data that is interleaved with frame data so that sounds
 * an be played as the movie is streamed over a network connection. 
 **/

void ASSound::processSoundStreamBloc(SoundStreamBlockTag *tag) {
	
	UI16 spriteId = tag->getCurSpriteId();
	setDecodingFinished(spriteId, false);
	timer->setSoundStreamPlaying(spriteId, PLAYING);

	// Initializing ffmpeg variables and testing for supported format
	// Makes a separate audioQueue for this sound stream and inserts it info the map audioQueues		
	audioQueues->newQueue(spriteId);

	int curFrameNumber = 0;
	int len = 0;
	
	SoundBlock* sb;
	SOUNDTRACK *sTrack = new SOUNDTRACK;
	sTrack->soundFormat = streamSoundCompression;
	sTrack->soundRate= streamSoundRate;
	sTrack->soundSize= streamSoundSize;
	sTrack->soundType= streamSoundType;
	sTrack->soundSampleCount= streamSoundSampleCount;

	// The mp3 decoders should only be opened once.
	// They are used in the function decodeSoundTrack, called below
	if(streamSoundCompression == 2){
		sTrack->pCodec=avcodec_find_decoder(CODEC_ID_MP3);
		if (!sTrack->pCodec) {
			fprintf(stderr, "MP3 codec not found in ffmpeg\n");
			return;
		}
		sTrack->pCodecCtx= avcodec_alloc_context();
		if (avcodec_open(sTrack->pCodecCtx, sTrack->pCodec) <0) { puts(" Could not open Codec"); return; }
	}

	while((tag != NULL)){

		curFrameNumber = tag->getFrameNumber();

		// To keep track of where in frame. The first audio is actually before the frame in mp3
		int skip = -tag->getSkipSize();
		
		sTrack->soundDataSize = (size_t) tag->getDataSize();	
		sTrack->soundData = tag->getStreamSoundData();

		while((sTrack->soundDataSize > 0) && (timer->getSoundStreamPlaying(spriteId) != EXITING)){
			
			len = decodeSoundTrack(sTrack, true);
			
			sTrack->soundDataSize -= len;
			sTrack->soundData = sTrack->ptrData;
			sb = sTrack->outputBuffers.front();

			// This must wait till the pointers are updated, 
			// otherwise decodeSoundTrack would be iven the same defunct frame over and over
			if (len < 0) {
				fprintf(stderr, "Error while decoding SOUND\n");
				delete sb;
				continue;
			}

			// It is vital that any mangled sound frames (that means, frames where the
			// outSize becomes negative) do not make it to the soundcard
			// There is written an error message in the deocdeSoundTrack function, so 
			// all we have to do here is discard them
			if(sb->dlen > 0){	
				sb->frame = tag->getFrameNumber();
				sb->skip = skip;
				skip += sb->dlen;

				// Adding soundblock to queue, and signaling there is more on the queue for fillAudioHandler
				streamstate State = timer->getSoundStreamPlaying(spriteId);
				if(State == EXITING){
					delete sb;
					delete sTrack;
					break;
				}
				soundInProgress = true;
				audioQueues->push(spriteId, sb);
				if(soundOpened)
					SDL_PauseAudio(0);
					
			} else
				delete sb;
	
			sTrack->outputBuffers.pop_front();
			assert(sTrack->outputBuffers.empty());

		} // while soundDataSize > 0

		soundInProgress = true;
		SoundStreamBlockTag *tmp = tag->getNextBlock();
		tag = tmp;

		
	} // while tag != NULL
	if(streamSoundCompression == 2){
		av_free(sTrack->pCodecCtx);
	}
	

	setDecodingFinished(spriteId, true);

	// Thread should not exit before all sound has been played or EXITING
	timer->waitForStreamEnd(spriteId);

	delete sTrack;
	
	timer->setSoundStreamPlaying(spriteId, FINISHED);
}


/**
 * Send a button sound the the sound hardware
 *
 * TODO: use SOUNDINFO properties
 **/
void ASSound::playButtonSound(int soundId, SOUNDINFO *p) {
	DefineSoundTag*t = soundTags[soundId];
	if (t != NULL) {
		registerSoundTrack(t);
		startSound(soundId,p);
	}
}


/**
 * Audio call back handler. 
 * It's C code and called asynchronously by the SDL lib.
 *
 **/
void fillAudioHandler(void *udata, Uint8 *stream, int len) {
	ASSound *so = (ASSound*) udata;
	if (!so->getSoundInProgress()) {
		return;
	}
	pthread_mutex_lock(&audioSynchro);	
	SwfTimer* timer = so->getTimer();
	bool noMoreData = false;

	AudioQueues* audioQueues = so->getAudioQueues();
        map<UI16,SoundBlock*> *curBlocks = new map<UI16,SoundBlock*>;
	int streamingSounds = audioQueues->size();	
  
        if(streamingSounds > 0){
	  // Fetching all audio queues = concurrent sound streams
	  
	  vector<UI16> *sprites = audioQueues->getSpriteIds();
	  vector<UI16>::iterator queueIt;	
	  vector<UI16>::iterator queueEnd;
  
	  for(queueIt = sprites->begin(); queueIt != sprites->end(); ++queueIt){
		  UI16 curSpriteId = *queueIt;
		  if(timer->getSoundStreamPlaying(curSpriteId) == EXITING){
			  audioQueues->erase(curSpriteId);
			  streamingSounds = audioQueues->size();
		  } else if(timer->getSoundStreamPlaying(curSpriteId) == PLAYING){
			  if(so->getDecodingFinished(curSpriteId) && audioQueues->empty(curSpriteId)){	
				  audioQueues->erase(curSpriteId);
				  streamingSounds = audioQueues->size();
				  timer->setSoundStreamPlaying(curSpriteId,EXITING);
			  } else {
				  (*curBlocks)[curSpriteId] = audioQueues->front(curSpriteId);
			  }
		  }
	  }
	  delete sprites;
	}
	
	while (len > 0  && !noMoreData) {

		int activeTracks= 0;
		int mergeTrackSize = 100000;
		noMoreData = true;
		
		// Here we find the length of the longest sound available by cheking all first streaming then event
		if(streamingSounds > 0){
			typedef map<UI16,SoundBlock*>::iterator blockIter;
			for(blockIter soundIt = curBlocks->begin(); soundIt != curBlocks->end(); soundIt++){
				SoundBlock* curBlock = soundIt->second;
				UI16 curSpriteId = soundIt->first;
				
				//This means that this stream should stop as fast as possible
				// EXITING is set by other functions wanting stream to stop
				if(timer->getSoundStreamPlaying(curSpriteId) == EXITING){
					curBlocks->erase(curSpriteId);
					streamingSounds = audioQueues->size();
					if(timer->getSoundStreamPlaying(curSpriteId) == PLAYING)
						timer->setSoundStreamPlaying(curSpriteId,EXITING);
					continue;
				}
				// If the current soundblock is empty or destroyed, pop it from queue
				if(curBlock->dlen <= 0){
					assert(audioQueues->size(curSpriteId) > 0);
					audioQueues->pop(curSpriteId);
					delete curBlock;
					// Check if stream is finished, in case remove stream
					if(so->getDecodingFinished(curSpriteId) && audioQueues->empty(curSpriteId)){
						curBlocks->erase(curSpriteId);
						streamingSounds = audioQueues->size();
						if(timer->getSoundStreamPlaying(curSpriteId) == PLAYING)
							timer->setSoundStreamPlaying(curSpriteId,EXITING);
						continue;
					}
				} // end if(curBlock->dlen <= 0)

				if(timer->getSoundStreamPlaying(curSpriteId) == PLAYING){
					(*curBlocks)[curSpriteId] = audioQueues->front(curSpriteId);
					// Update info about sound
					int restStream = curBlock->dlen;
					if(restStream >= 0){
						int len1 = restStream > (unsigned int) len ? len : restStream ;
						mergeTrackSize = mergeTrackSize > len1 ? len1 : mergeTrackSize;
						activeTracks ++;
						noMoreData = false;
					} else {
						wcout << "fillAudioHandler: Corrupted audio frame: negative length\n";
					}
					continue;	
				} // end if
			} // end for
			
		} // end if (streamingSound > 0)
		
		// Update info about event sounds
		for (IntSOUNDTRACKHashMap::const_iterator it = (so->getSoundTracks())->begin(); it != (so->getSoundTracks())->end(); it++) {
                        
			SOUNDTRACK *s = (*it).second;
			if (s->outputBuffers.size() < 20 ) {
				so->decodeNextBlock(s);
				so->decodeNextBlock(s);
			}

			if (s->outputBuffers.size() == 0) continue;
			SoundBlock * sb = *(s->outputBuffers.begin());
			if (sb->dlen == 0 || sb->buffer == NULL) {
				delete sb;
				s->outputBuffers.pop_front();
			} else {
				int len1 = sb->dlen > (unsigned int)len ? len : sb->dlen ;
				mergeTrackSize = mergeTrackSize > len1 ? len1 : mergeTrackSize;
				activeTracks ++;
				noMoreData = false;	
			}
		}
		
		// This is where the actual mixing takes place. First streaming, then event is added to mergeTrack
		if (mergeTrackSize > 0 && activeTracks > 0) {
			UI8 *mergeTrack = (UI8*)calloc(mergeTrackSize,1);
			
			// First get out all streaming sound, if any
			if(streamingSounds > 0){
				typedef map<UI16,SoundBlock*>::iterator blockIter;
				for(blockIter soundIt = curBlocks->begin(); soundIt != curBlocks->end(); soundIt++){
					UI16 curSpriteId = soundIt->first;
					// If signal is given for soundstream to stop, back out
					if(timer->getSoundStreamPlaying(curSpriteId) != PLAYING && audioQueues->empty(curSpriteId)){
						curBlocks->erase(curSpriteId);	
						continue;
					}
					SoundBlock* curBlock = soundIt->second;
					assert (curBlock->dlen >= mergeTrackSize);
					UI8 *p = (UI8*)(&curBlock->buffer[curBlock->dpos]);
					for (int i=0; i < mergeTrackSize; i++) {
						int c = mergeTrack[i] + p[i];
						if (c > 255) c = 255;
						mergeTrack[i] = c;
					}
					curBlock->dpos += mergeTrackSize;
					curBlock->dlen -= mergeTrackSize;
					noMoreData = false;

					// Essential for audio syncing. The video part is set by SwfEngine - case stagSoundStreamBlock
					// spriteId = 0 means this is audio from the main timeline
					timer->setSoundFrame(curSpriteId, curBlock->frame, curBlock->skip);
				}
			}
                       
			for (IntSOUNDTRACKHashMap::const_iterator it = (so->getSoundTracks())->begin(); it != (so->getSoundTracks())->end(); it++) {
				SOUNDTRACK *s = (*it).second;
				if (s->outputBuffers.size() == 0) continue;
				SoundBlock * sb = *(s->outputBuffers.begin());
				UI8 *p = (UI8*)(&sb->buffer[sb->dpos]);
				for (int i=0; i< mergeTrackSize; i++) {
					int c = mergeTrack[i] + p[i];
					if (c>255) c = 255;
					mergeTrack[i] = c;
				}
				sb->dpos += mergeTrackSize;
				sb->dlen -= mergeTrackSize;
				noMoreData = false;
				if (sb->dlen == 0) {
					delete sb;
					sb = NULL;
					s->outputBuffers.pop_front();
				}
			}
			memcpy(stream,mergeTrack,mergeTrackSize);         
			len -= mergeTrackSize;
			stream += mergeTrackSize;
			free(mergeTrack);
		} // end if (mergeTrackSize > 0 && activeTracks > 0)
	} // end while (len > 0 && !noMoreData)
	
	if (noMoreData && audioQueues->empty()) {
		so->setSoundInProgress(false);
	}
	delete curBlocks;
	pthread_mutex_unlock(&audioSynchro);
}

/**
 * FFMPEG can't decode SWF ADPCM without crashing. We provide here an alternate decoding lib.
 * 
 * The decoding code is here
 **/
// This file has been rearranged from the code posted
// on news:forums.macromedia.com by Jonathan Gay.
// Courtesy of Macromedia

//
// ADPCM tables
//

static const int indexTable2[2] = {
    -1, 2,
};

// Is this ok?
static const int indexTable3[4] = {
    -1, -1, 2, 4,
};

static const int indexTable4[8] = {
    -1, -1, -1, -1, 2, 4, 6, 8,
};

static const int indexTable5[16] = {
 -1, -1, -1, -1, -1, -1, -1, -1, 1, 2, 4, 6, 8, 10, 13, 16,
};

static const int* indexTables[] = {
 indexTable2,
 indexTable3,
 indexTable4,
 indexTable5
};

static const int stepsizeTable[89] = {
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

long Adpcm::getBits(int n) {
	if ( bitPos < n ) fillBuffer();
	long v = ((UI32)bitBuf << (32-bitPos)) >> (32-n);
	bitPos -= n;
	return v;
}

long Adpcm::getSBits(int n) {
	if ( bitPos < n ) fillBuffer();
	long v = ((SI32)bitBuf << (32-bitPos)) >> (32-n);
	bitPos -= n;
	return v;
}

// Constructor
Adpcm::Adpcm(unsigned char *buffer, long isStereo) {
	stereo = isStereo;
	src = buffer;

	nBits = 0; // flag that it is not inited
	nSamples = 0;

	bitPos = 0;
	bitBuf = 0;
}

void Adpcm::fillBuffer() {
	while ( bitPos <= 24 ) {
		bitBuf = (bitBuf<<8) | *src++;
		bitPos += 8;
	}
}

int Adpcm::decompress(short *dst, long n) {
	short *dst0 = dst;
	if ( nBits == 0 ) {
		// Get the compression header
		nBits = (int)getBits(2)+2;
	}

	const int* indexTable = indexTables[nBits-2];
	int k0 = 1 << (nBits-2);
	int signmask = 1 << (nBits-1);

	if ( !stereo ) {
		// Optimize for mono
		long	vp = valpred[0]; // maybe these can get into registers...
		int		ind = index[0];
		long	ns = nSamples;

		while ( n-- > 0 ) {
			ns++;

			if ( (ns & 0xFFF) == 1 ) {
				// Get a new block header
				*dst++ = (short)(vp = getSBits(16));

				ind = (int)getBits(6); // The first sample in a block does not have a delta
			} else {
				// Process a delta value
				int delta = (int)getBits(nBits);

				// Compute difference and new predicted value
				// Computes 'vpdiff = (delta+0.5)*step/4'
				int step = stepsizeTable[ind];
				long vpdiff = 0;
				int k = k0;

				do {
					if ( delta & k )
					vpdiff += step;
					step >>= 1;
					k >>= 1;
				} while ( k );

				vpdiff += step; // add 0.5

				if ( delta & signmask ) // the sign bit
					vp -= vpdiff;
				else
					vp += vpdiff;

				// Find new index value
				ind += indexTable[delta&(~signmask)];

				if ( ind < 0 )
					ind = 0;
				else if ( ind > 88 )
					ind = 88;

				// clamp output value
				if ( vp != (short)vp )
					vp = vp < 0 ? -32768 : 32767;

				/* Step 7 - Output value */
				*dst++ = (short)vp;
			}
		}

		valpred[0] = vp;
		index[0] = ind;
		nSamples = ns;

	} else {
		int sn = stereo ? 2 : 1;

		// Stereo
		while ( n-- > 0 ) {
			nSamples++;
			if ( (nSamples & 0xFFF) == 1 ) {
				// Get a new block header
				for ( int i = 0; i < sn; i++ ) {
					*dst++ = (short)(valpred[i] = getSBits(16));
					// The first sample in a block does not have a delta
					index[i] = (int)getBits(6);
				}
			} else {
				// Process a delta value
				for ( int i = 0; i < sn; i++ ) {
					int delta = (int)getBits(nBits);

					// Compute difference and new predicted value
					// Computes 'vpdiff = (delta+0.5)*step/4'
					int step = stepsizeTable[index[i]];
					long vpdiff = 0;
					int k = k0;

					do {
						if ( delta & k ) vpdiff += step;
						step >>= 1;
						k >>= 1;
					} while ( k );
					vpdiff += step; // add 0.5

					if ( delta & signmask ) // the sign bit
						valpred[i] -= vpdiff;
					else
						valpred[i] += vpdiff;

					// Find new index value
					index[i] += indexTable[delta&(~signmask)];

					if ( index[i] < 0 )
						index[i] = 0;
					else if ( index[i] > 88 )
						index[i] = 88;

					// clamp output value
					if ( valpred[i] != (short)valpred[i] )
						valpred[i] = valpred[i] < 0 ? -32768 : 32767;

					/* Step 7 - Output value */
					*dst++ = (short)valpred[i];
				}
			}
		}
	}
	return (dst-dst0)*2;
}


/**
 * AudioQueues is a wrapper around the map with the streaming sound
 * It has reesponsibility for preventing concurrent access from several threads
 * and for freeing and allocating memory
*/
AudioQueues::AudioQueues(){
	wcout << "Error: Standard constructor AudioQueues::AudioQueues()\n";
	exit(1);
}


AudioQueues::AudioQueues(SwfTimer *atimer){
	timer = atimer;
	queueLocks = 0;
}

AudioQueues::~AudioQueues(){
	// Removes all streaming sounds
	lockQueue();
	typedef map<UI16,queue<SoundBlock*> >::const_iterator iter;
	for(iter strIt = audioQueues.begin(); strIt != audioQueues.end(); strIt++){
		queue<SoundBlock*> audioQueue = audioQueues[strIt->first];
		while(!audioQueue.empty()){
			audioQueue.pop();
		}
	}
	// Must now signal any threads waiting in push() or front() to exit before destructor exiting
	pthread_cond_broadcast(&fullStream);
	pthread_cond_broadcast(&emptyStream);
	
	while(queueLocks > 0)
		pthread_cond_wait(&putWaiting, &queueSynchro);

	unlockQueue();
}

/**
 * Returns vector with ids of all sprites with currently running streaming sounds
 * Most of the operations do the operation with the same name on the map representing 
 * the audio. In addition everuthing is threadsafe.
*/
vector<UI16>* AudioQueues::getSpriteIds(){
	lockQueue();
	map<UI16,queue<SoundBlock*> >::const_iterator iter;
	vector<UI16>* sprites = new vector<UI16>(audioQueues.size());
	for(iter = audioQueues.begin(); iter != audioQueues.end(); iter++){
		sprites->push_back(iter->first);
	}
	pthread_mutex_unlock(&queueSynchro);
	return sprites;
}	

void AudioQueues::newQueue(UI16 spriteId){
	lockQueue();
	audioQueues[spriteId] = *(new queue<SoundBlock*>);
	unlockQueue();
}

void AudioQueues::push(UI16 spriteId, SoundBlock* sb){

	lockQueue();
	queueLocks++;
	if(audioQueues.find(spriteId) == audioQueues.end()){
		wcout << "Trying to push to non-existing audioQueue\n";
		exit(1);
	}
	queue<SoundBlock*> *audioQueue = &(audioQueues[spriteId]);
	audioQueue->push(sb);
	pthread_cond_signal(&emptyStream);
	while(audioQueue->size() > 4 && timer->getSoundStreamPlaying(spriteId) == PLAYING){
		pthread_cond_wait(&fullStream, &queueSynchro);
		audioQueue = &(audioQueues[spriteId]);
	}
	queueLocks--;
	unlockQueue();
	//Signal destructor to exit if waiting
	pthread_cond_signal(&putWaiting);

}

/**
 * Signals any sleeping threads
*/

void AudioQueues::signal(){
	lockQueue();
	pthread_cond_broadcast(&fullStream);
	pthread_cond_broadcast(&emptyStream);
	unlockQueue();
}

/**
 * Pops soundblock from queue.
 * Frees bound memory in the block
*/
void AudioQueues::pop(UI16 spriteId){
	lockQueue();
	if(audioQueues.find(spriteId) == audioQueues.end()){
		wcout << "Trying to pop non-existing audioQueue\n";
		exit(1);
	}
	queue<SoundBlock*> *audioQueue = &(audioQueues[spriteId]);
	if(audioQueue->empty()){
		wcout << "Trying to pop empty audioQueue " << spriteId << endl;
		exit(1);
	}
	audioQueue->pop();				

	// Tells any waiting thread in processSoundStreamBloc to continue decoding
	// (Since the queue is now not "full" anymore)
	pthread_cond_broadcast(&fullStream);
	unlockQueue();
}
	

SoundBlock* AudioQueues::front(UI16 spriteId){
	lockQueue();
	queueLocks++;
	if(audioQueues.find(spriteId) == audioQueues.end()){
		wcout << "Trying to get front of non-existing audioQueue\n";
		pthread_mutex_unlock(&queueSynchro);
		return NULL;
	}
	queue<SoundBlock*> *audioQueue = &(audioQueues[spriteId]);
	while(audioQueue->empty() && timer->getSoundStreamPlaying(spriteId) == PLAYING){
		pthread_cond_wait(&emptyStream, &queueSynchro);
		audioQueue = &(audioQueues[spriteId]);
	}
	SoundBlock* retBlock = NULL;
	if(timer->getSoundStreamPlaying(spriteId) == PLAYING)
		retBlock = audioQueue->front();
	queueLocks--;
	unlockQueue();
	pthread_cond_signal(&putWaiting);
	return retBlock;
	
}

void AudioQueues::erase(UI16 spriteId){
	lockQueue();
	
	if(audioQueues.find(spriteId) == audioQueues.end()){
		wcout << "Trying to remove non-existing audioQueue\n";
		pthread_mutex_unlock(&queueSynchro);
		return;
	}
	queue<SoundBlock*> audioQueue = audioQueues[spriteId];
	while(!audioQueue.empty()){
		audioQueue.pop();
	}
	audioQueues.erase(spriteId);
	unlockQueue();

}

void AudioQueues::clear(){
	// Removes all streaming sounds
	lockQueue();
	typedef map<UI16,queue<SoundBlock*> >::iterator iter;
	for(iter strIt = audioQueues.begin(); strIt != audioQueues.end(); strIt++){
		queue<SoundBlock*> audioQueue = audioQueues[strIt->first];
		while(!audioQueue.empty()){
			delete ((audioQueue.front())->buffer);
			audioQueue.pop();
		}
	}
	audioQueues.clear();
	unlockQueue();
}

bool AudioQueues::empty(){
	lockQueue();
	bool empty = audioQueues.empty();
	unlockQueue();
	return empty;
}

bool AudioQueues::empty(UI16 curSpriteId){
	lockQueue();
	if(audioQueues.find(curSpriteId) == audioQueues.end()){
		wcout << "Trying to test if non-existing audioQueue is empty\n";
		pthread_mutex_unlock(&queueSynchro);
		return true;
	}
	bool empty = audioQueues[curSpriteId].empty();
	unlockQueue();
	return empty;
}

int AudioQueues::size(){
	lockQueue();
	int size = audioQueues.size();
	unlockQueue();
	return size;
}


int AudioQueues::size(UI16 spriteId){
	lockQueue();
	if(audioQueues.find(spriteId) == audioQueues.end()){
		wcout << "Trying to find size of non-existing audioQueue\n";
		exit(1);
	}
	int size = audioQueues[spriteId].size();
	unlockQueue();
	return size;
}

void AudioQueues::lockQueue(){
	int err = pthread_mutex_lock(&queueSynchro);
	if(err != 0){
		wcout << "mutex queueSynchro not properly initialized or repeated locking of queueSynchro\nError # " << err << endl;
		exit(1);
	}
}

void AudioQueues::unlockQueue(){
	int err = pthread_mutex_unlock(&queueSynchro);
	if(err != 0){
		wcout << "mutex queueSynchro unlock failing (impossible?)\n";
		exit(1);
	}
}
