/**
Macromedia Flash File graphic lib.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "bgraphics.hh"
#include "localFonts.hh"
#include <list>
#include <stack>
#include "ASLib/as_text_field.hh"
#include "ASLib/as_text_format.hh"

/**
 * The constructor. This class is abstract - can't be instanciated.
 **/
Graphics::Graphics(BaseTexture *texture, LocalFontManager *localFontManager) { 
	points= NULL;
	segments=NULL;
	fillStyles=NULL;
	isCharacter=false;
	raster=NULL;
	this->texture = texture;
	this->localFontManager = localFontManager;
	allowLowPrecision = false;
}

/**
 * The destructor
 **/
Graphics::~Graphics() {
	if (points != NULL) { delete []points; points = NULL; }
	if (segments != NULL) { delete []segments; segments = NULL; }
	if (fillStyles != NULL) { delete []fillStyles; fillStyles = NULL; }
	if (raster != NULL) { delete []raster; raster = NULL; }
	for (DrawingSequenceEntries::iterator it=drawingSequences.begin();it != drawingSequences.end(); it++) {
		ListDrawingSequence *l = it->second;
		// If we have an other entry with same listDrawingSequence, remove if.
		for (DrawingSequenceEntries::iterator it2=drawingSequences.begin(); it2 != drawingSequences.end(); ++it2) {
			if (it != it2 && it2->second == it->second) drawingSequences.erase(it2);
		}

		if (l != NULL) {
			clearSequence(l);
			l->clear();
			delete l;
		}
	}
	drawingSequences.clear();
	this->texture = NULL;
	this->localFontManager = NULL;
}

/**
 * Estimate max number of points and segments to enable allocation of working data variables
 * Due to flush commands, this is probably too much memory but not that big.
 **/
int Graphics::getShapeAllocSize(Shape *shape) {
	int size = 1;
	list<ShapeRecord*>::iterator it;
	for (it = shape->getShapeRecords()->begin(); it != shape->getShapeRecords()->end(); it++) {
		UI8 type = (*it)->getTypeFlag();
		switch (type) {
			case StyleChangeRecordType:	size ++; break;
			case StraightEdgeRecordType: allowLowPrecision ? size+=6 : size++; break;
			case CurvedEdgeRecordType: size += 6; break;
		}
	}
	return size;
}

/**
 * Insert a point in the mesh for triangulation 
 * 
 * If point exist return its location, else create a new entry.
 **/
int Graphics::findInsertPoint(float x, float y) {
	for (int i=0; i< nbPoints; i++) {
		if (allowLowPrecision && fabs(points[i*2]-x)<10.0 && fabs(points[i*2+1]-y)<10.0) return i;
		if (!allowLowPrecision && points[i*2]==x && points[i*2+1] == y) return i;
	}
	points[nbPoints*2]=x; 
	points[nbPoints*2+1]= y;
	nbPoints++;
	return nbPoints-1;
}

/**
 * Insert a Color in the mesh for triangulation 
 *
 * Color indexes are stored in the mesh during triangulation. 
 * They are then used to convert color indexes to RGBA.
 **/
int Graphics::findInsertColor(FILLSTYLE *c) {
	for (int i=0; i< nbFillStyles; i++) {
		if (c == fillStyles[i])	return i;
	}
	fillStyles[nbFillStyles] = c;
	nbFillStyles++;
	return nbFillStyles-1;
}

/**
 * Allocate the memory objects required for processing a Shape
 **/
void Graphics::allocateShapeProcessObjects(Shape *shape) {
	int size= getShapeAllocSize(shape);
	points = new float[size*2];
	segments = new Segment[size+1]; // extra space for extra closing segments if needed
	fillStyles = new FILLSTYLE*[1024];
	fillStyles[0] = &defaultFillStyle;
	nbPoints=0;
	nbSegments=0;
	nbFillStyles=1;
}

/**
 * creates the computed texture (linear/gradient) if needed
 **/
void Graphics::registerComputedTexture(int fillStyleType,int fillStyleIndex, FILLSTYLE *fill ) {
	if (fill != NULL) {
		switch (fillStyleType) {
			case 0x0: 
				texture->registerColor(fill[0].getColor());
				break;
			case 0x10: {
				int texBuf = texture->getGradient(characterId,fillStyleIndex);
				if (texBuf == 0) texture->registerLinearGradient(characterId,fillStyleIndex,fill);
			break;
			}
			case 0x12: {
				int texBuf = texture->getGradient(characterId,fillStyleIndex);
				if (texBuf == 0) texture->registerRadialGradient(characterId,fillStyleIndex,fill);
			break;
			}
		}
	}
}

/**
 * Update Segment with all values
 * Check for overlapping segments.
 * When segments overlap, this cause color interior fill to be defined twice or more and we get wrong color.
 **/
void Graphics::updateSegments() {
	for (int i=0; i< nbSegments; i++) {
		segments[i].x[0] = points[segments[i].points[0]*2];
		segments[i].x[1] = points[segments[i].points[1]*2];
		segments[i].y[0] = points[segments[i].points[0]*2+1];
		segments[i].y[1] = points[segments[i].points[1]*2+1];
	}
}

/**
 * Init configuration to raster a character definition
 **/
void Graphics::currentShapeIsCharacter() {
	raster = new UI8[64*64];
	memset(raster,0,64*64); // The shape may be empty, in case of space char
	isCharacter = true;
}

/**
 * It goes to the shape, analyze the shape records, the fill styles, new fillstyles and creates a list of segments.
 * The segments are passed to the triangulator or the rasterizer.
 * For better performance & look:
 * - the character are sent to the rasterizer to allow antialiasing with textures (GL_LINEAR filtering),
 * - the shapes are passed to the rasterizer because they tends to be rectangles, textured mapped geometrical objects.
 *
 **/
void Graphics::processShape(Shape *shape,int characterId,RECT * rect) {
	//printf("Process ident:%x  Shape:%x\n",characterId,shape);
	int prevPoint=0;
	int	newPoint;
	int nbContours=0;
	float x=0,y=0;
	StyleChangeRecord *sDef = NULL;
	ShapeWithStyle *s2 = NULL; 
	int originalFillStyleIndex0=-1, originalFillStyleIndex1=-1;
	int colorFillStyleIndex0=-1, colorFillStyleIndex1=-1;
	FILLSTYLE *fill = NULL;
	this->characterId = characterId;
	list<ShapeRecord*>::const_iterator it;
	this->rect = rect;
	int  nb_edges=0;
	LINESTYLE *lineColor=NULL;
	allocateShapeProcessObjects(shape);
	if (shape->getIsShapeWithStyle()) s2 = (ShapeWithStyle*)shape;
	bool hasTriangles = false;
	bool hasLines = false;

	for (it = shape->getShapeRecords()->begin(); it != shape->getShapeRecords()->end(); it++) {
		UI8 type = (*it)->getTypeFlag();
		switch (type) {
			case FlushShapeRecordType: {
				if (hasTriangles) insertSequence(characterId,shape2Triangles());
				if (hasLines) insertSequence(characterId,shape2Lines());
				prevPoint = newPoint= nbContours = 	nbSegments = nbPoints = 0;
				break;
			}
			case StyleChangeRecordType:	{
				StyleChangeRecord *s = (StyleChangeRecord*) *it;
				if (s->getStateNewStyles()) {
					 sDef = s; 
				}

				if (s->getStateFillStyle0()) {
					int idx = s->getFillStyle0();
					if (idx > 0) {
						if (sDef != NULL) {
							fill = &sDef->getFillStyles()[idx-1];
						} else {
							if (s2 != NULL) fill = &s2->getFillStyles()[idx-1];
						}
						if (fill!=NULL) {
							originalFillStyleIndex0 = idx-1;
							colorFillStyleIndex0 = findInsertColor(fill);
							registerComputedTexture(fill->getFillStyleType(),colorFillStyleIndex0,fill);
						}
						hasTriangles = true;
					} else {
						originalFillStyleIndex0 = -1;
					}			
				}

				if (s->getStateFillStyle1()) {
					int idx = s->getFillStyle1();
					if (idx > 0) {
						if (sDef != NULL) {
							 fill = &sDef->getFillStyles()[idx-1];
						} else {
							if (s2 != NULL) fill = &s2->getFillStyles()[idx-1];
						}
						if (fill!=NULL) {
							originalFillStyleIndex1 = idx-1;
							colorFillStyleIndex1 = findInsertColor(fill);
							registerComputedTexture(fill->getFillStyleType(),colorFillStyleIndex1,fill);
						}
						hasTriangles = true;
					} else {
						originalFillStyleIndex1 = -1;
					}
				}

				if (s->getStateLineStyle()) {
					int idx = s->getLineStyle();
					if (idx > 0) {
						if (sDef != NULL) {
							lineColor = &sDef->getLineStyles()[idx-1];
						} else {
							if (s2 != NULL) {
								lineColor = &s2->getLineStyles()[idx-1];
							}
						}
						hasLines = true;
					} else {
						lineColor= NULL;
					}
				} 

				if (s->getStateMoveTo()) {
					if (nb_edges == 0 && !isCharacter) {			// Flush current path
						if (nbSegments > 0 &&nbPoints>0) {
							updateSegments();
							bool isClosed = true; //LOOKS BUGGY IN FACT: isClosedPolygon(segments,nbSegments,nbPoints);
							if (hasTriangles && isClosed) insertSequence(characterId,shape2Triangles());
							if (hasLines) insertSequence(characterId,shape2Lines());
							prevPoint = newPoint= nbContours = 	nbSegments = nbPoints = 0;
						}
					}
					x = s->getMoveDeltaX();
					y = s->getMoveDeltaY();
					prevPoint = findInsertPoint(x,y);  // test Buggy if (x!=0 && y!=0) 
					nbContours++;
					nb_edges = 0;
				}
				break; 
			}
			case StraightEdgeRecordType: {
				StraightEdgeRecord *s = (StraightEdgeRecord*) *it;
				nb_edges++;
				float nx = x, ny = y;
				if (s->getGeneralLineFlag()) {
					nx += s->getDeltaX();
					ny += s->getDeltaY();
				} else {
					if (s->getVertLineFlag()) {
						ny += s->getDeltaVert();
					} else {
						nx += s->getDeltaVert();
					}
				}
				x = nx ; y = ny;
				newPoint = findInsertPoint(x,y);
				if (prevPoint != newPoint) {
					Segment *s = (Segment*) &segments[nbSegments];
					s->points[0] = prevPoint;
					s->points[1] = newPoint;
					if (points[prevPoint*2+1] < points[newPoint*2+1]) {
						s->originalIndex[0] = originalFillStyleIndex0;
						s->originalIndex[1] = originalFillStyleIndex1;
						s->colorsIndex[0] = colorFillStyleIndex0;
						s->colorsIndex[1] = colorFillStyleIndex1;
					} else {
						s->originalIndex[0] = originalFillStyleIndex1; 
						s->originalIndex[1] = originalFillStyleIndex0;
						s->colorsIndex[0] = colorFillStyleIndex1;
						s->colorsIndex[1] = colorFillStyleIndex0;
					}
					s->lineColor = lineColor;
					nbSegments++;
					prevPoint = newPoint;
				}
				break; 
			}
			case CurvedEdgeRecordType: {
				CurvedEdgeRecord *s = (CurvedEdgeRecord*) *it;
				nb_edges++;
				float cx = x + s->getControlDeltaX();
				float cy = y + s->getControlDeltaY();
				float ex = cx + s->getAnchorDeltaX();
				float ey = cy + s->getAnchorDeltaY();
				for (int i=0; i < 5; i++) {
					float t = (float)i * 0.2f;
					float curX = QuadraticBezierInterp(x,cx,ex,t);
					float curY = QuadraticBezierInterp(y,cy,ey,t);
					newPoint = findInsertPoint(curX,curY);
					if (prevPoint != newPoint) {
						Segment *s = (Segment*) &segments[nbSegments];
						s->points[0] = prevPoint;
						s->points[1] = newPoint;
						if (points[prevPoint*2+1] < points[newPoint*2+1]) {
							s->originalIndex[0] = originalFillStyleIndex0;
							s->originalIndex[1] = originalFillStyleIndex1;
							s->colorsIndex[0] = colorFillStyleIndex0;
							s->colorsIndex[1] = colorFillStyleIndex1;
						} else {
							s->originalIndex[0] = originalFillStyleIndex1; 
							s->originalIndex[1] = originalFillStyleIndex0;
							s->colorsIndex[0] = colorFillStyleIndex1;
							s->colorsIndex[1] = colorFillStyleIndex0;
						}
						s->lineColor = lineColor;
						nbSegments++;
						prevPoint = newPoint;
					}
				}
				x = ex;
				y = ey;
				newPoint = findInsertPoint(x,y);
				if (prevPoint != newPoint) {
					Segment *s = (Segment*) &segments[nbSegments];
					s->points[0] = prevPoint;
					s->points[1] = newPoint;
					if (points[prevPoint*2+1] < points[newPoint*2+1]) {
						s->originalIndex[0] = originalFillStyleIndex0;
						s->originalIndex[1] = originalFillStyleIndex1;
						s->colorsIndex[0] = colorFillStyleIndex0;
						s->colorsIndex[1] = colorFillStyleIndex1;
					} else {
						s->originalIndex[0] = originalFillStyleIndex1; 
						s->originalIndex[1] = originalFillStyleIndex0;
						s->colorsIndex[0] = colorFillStyleIndex1;
						s->colorsIndex[1] = colorFillStyleIndex0;
					}
					s->lineColor = lineColor;
					nbSegments++;
					prevPoint = newPoint;
				}
				break; 
			}
		}
	}
	if (nbContours == 0 && nbPoints>0 ) {	// Force closing the figure, so we have a closed area
		Segment *s = (Segment*) &segments[nbSegments];
		s->points[0] = prevPoint;
		s->points[1] = 0;
		if (points[prevPoint*2+1] < points[1]) {
			s->originalIndex[0] = originalFillStyleIndex0;
			s->originalIndex[1] = originalFillStyleIndex1;
			s->colorsIndex[0] = colorFillStyleIndex0;
			s->colorsIndex[1] = colorFillStyleIndex1;
		} else {
			s->originalIndex[0] = originalFillStyleIndex1; 
			s->originalIndex[1] = originalFillStyleIndex0;
			s->colorsIndex[0] = colorFillStyleIndex1;
			s->colorsIndex[1] = colorFillStyleIndex0;
		}
		s->lineColor = lineColor;
		nbSegments++;
	}
	if (nbSegments > 0 &&nbPoints>0) {
		updateSegments();
		bool isClosed = true; // isClosedPolygon(segments,nbSegments,nbPoints);
		//closePolygon(segments,nbSegments,points,nbPoints);
		if (isCharacter) {
			shape2Bitmap(); 
		} else {
			if (hasTriangles && isClosed) insertSequence(characterId,shape2Triangles());
			if (hasLines)	insertSequence(characterId,shape2Lines());
		}
	}
	delete []points; points=NULL;
	delete []segments; segments=NULL;
	delete []fillStyles; fillStyles=NULL;
}


/**
 * Convert a Shape to a texture (Character of type non local).
 * Local characters are converted into bitmaps by true type lib.
 *
 * The swf fill format is complex with overlapping areas and holes.
 *
 **/
void Graphics::shape2Bitmap() {
	UI8* tmpRes = new UI8[64*64*4];
	memset(tmpRes,0,64*64*4);
	for (int yi=0; yi< 128; yi++) {
		list<int> intersect;
		list<float> intersectX;
		float	xmin = -1;
		int	  	color=0;
		float yt = 255.5-yi*9;
		int	  lastX, lastX2;

		// Get list of intersect
		for (int i=0; i< nbSegments; i++) {
			Segment *s = &segments[i];
			if (s->y[0] > yt && s->y[1] > yt) continue;
			if (s->y[0] < yt && s->y[1] < yt) continue;
			float dy = s->y[1] - s->y[0];
			if (dy == 0) continue;
			float t = (yt-s->y[0])/dy;
			if (t<0 && t>1) continue;
			float dx = s->x[1] - s->x[0];
			float xi = s->x[0] + t*dx + 128;
			
			bool doInsert = true;
			for (list<float>::iterator it = intersectX.begin(); it != intersectX.end() && doInsert; it++) {
				if (*it == xi) {
					doInsert = false;				
				}
			}
			if (doInsert) {
				intersect.push_back(i);
				intersectX.push_back(xi);
			}
		}

		while (xmin < 1152) {
			list<int>::iterator it;
			list<float>::iterator it2;
			float newxmin = 10000;
			for (it = intersect.begin(), it2 = intersectX.begin(); it != intersect.end(); it++, it2++) {
				if (*it2 > xmin && newxmin>*it2) {
					newxmin=*it2;
				}		
			}
			if (newxmin < 1152) {
				for (int a=int(xmin); a < int(newxmin) && a<1152; a ++) {
					if (a>=0) tmpRes[(127-yi)*128+int(a/9.0f)] = color;
				}
				lastX2 = int(lastX);
				lastX = int(newxmin);
				color = color == 0 ? 255 : 0;
			}
			xmin = newxmin+0.1;
		}
	}

	// Process final texture
	for (int i=0; i< 64; i++) {
		for (int j=0; j<64; j++) {
			int k = tmpRes[i*256+j*2] + tmpRes[i*256+j*2+1] + tmpRes[i*256+j*2+128] + tmpRes[i*256+j*2+1+128];
			raster[i*64+j] = k / 4;
			//raster[i*64+j] = tmpRes[i*256+j*2];
			//printf("%c",raster[i*64+j]>0 ? '1' : '0');
		}
		//puts("");
	}
	//puts("");fflush(stdout);
	delete []tmpRes;
}


/**
 * Convert drawing orders to pseudo display lists
 **/
DrawingSequence *Graphics::shape2Lines() {
	//puts("In shape2Lines");
	DrawingSequence	*drawingSequence = new DrawingSequence();
	drawingSequence->allocateVertices(2*nbPoints);
	float *fpoints = drawingSequence->getVertices();
	for (int i=0; i< nbPoints*2; i++) fpoints[i] = float(points[i]);
	drawingSequence->setNbPoints(nbPoints);
	// Register at least one color to create the color texture if necessary
	RGBA c; c.red=c.green=c.blue=c.alpha=127; texture->registerColor(&c);

	// Order the drawing by line width to optimize rendering.
	IntIntHashMap width;
	for (int i=0; i< nbSegments;i++) {
		if (segments[i].lineColor != NULL) {
			width[segments[i].lineColor->width] = segments[i].lineColor->width;
		}
	}

	// Loop over line width
	for (IntIntHashMap::iterator it = width.begin(); it != width.end(); it++) {
		int nbLines = 0;
		for (int i=0; i< nbSegments;i++) 
			if (segments[i].lineColor != NULL && segments[i].lineColor->width==width[(*it).first]) nbLines++;

		DrawingEntry *drawingEntry = new DrawingEntry();
		drawingEntry->allocateIndexes(2*nbLines);
		unsigned int *indexes = drawingEntry->getIndexes();
		drawingEntry->allocateTexCoords(2*nbPoints);
		float *texCoords = drawingEntry->getTexCoords();
		int p = 0;
		for (int i=0; i< nbSegments;i++) {
			if (segments[i].lineColor == NULL || segments[i].lineColor->width!=width[(*it).first])  continue;
			indexes[p*2+0] = segments[i].points[0];
			indexes[p*2+1] = segments[i].points[1];
			LINESTYLE *lineColor = segments[i].lineColor;
			int loc;
			if (lineColor != NULL) {
				texture->registerColor(&lineColor->color);
			 	loc = texture->getColorTextureIndex(&lineColor->color);
			} else {
				loc = 0; // Black with alpha = 0;
			}
			float tx = (loc%256)/256.0f;
			float ty = float(loc/256)/256.0f;
			texCoords[segments[i].points[0]*2+0] = tx;
			texCoords[segments[i].points[0]*2+1] = ty;
			texCoords[segments[i].points[1]*2+0] = tx;
			texCoords[segments[i].points[1]*2+1] = ty;
			p++;
		}
		drawingEntry->drawingType = DrawingLines;
		drawingEntry->fillType = 0;
		drawingEntry->texNum = texture->getColorTexture();
		drawingEntry->lineWidth = float(width[(*it).first])*0.05;
		drawingEntry->nbElements = nbLines;
		drawingSequence->addDrawingEntry(drawingEntry);
	}
	return drawingSequence;
}

/**
 * Create the rendering commands to display a text string
 **/
void Graphics::registerStaticString(long unid, DefineTextTag* tag) {
//puts("registerStaticString");
	list<TEXTRECORD*>::iterator it;
	int 	fontId = 0;
	float 	fHeight = 10;
	int 	prevTextureID = -1;
	DrawingSequence	*drawingSequence = new DrawingSequence();

	// Get Number of glyphs and alloc memory
	int size = 0;
	for (it = tag->getTextRecords()->begin(); it != tag->getTextRecords()->end(); it++) size += (*it)->getGlyphCount();
	nbPoints = 4*size;
	drawingSequence->allocateVertices(2*nbPoints);
	float *fpoints = drawingSequence->getVertices();
	float *texCoords = NULL;
	unsigned int *indexes = NULL;
	int iPoint = 0;
	int iSeg = 0;
	DrawingEntry *drawingEntry = new DrawingEntry();

	for (it = tag->getTextRecords()->begin(); it != tag->getTextRecords()->end(); it++) {
		drawingEntry->drawingType = DrawingFontChars;
		drawingEntry->fillType = 0;
		float dx= 0;
		float yoffset=0, xoffset=0;
		TEXTRECORD *r = *it;
		if (r->getStyleFlagsHasColor()) {
			RGBA c = r->getTextColor();
			if (tag->getType() == stagDefineText) c.alpha = 255;
			drawingEntry->charColor = c;
		}
		if (r->getStyleFlagsHasFont()) 		fHeight = float(r->getTextHeight());
		if (r->getStyleFlagsHasFont()) 		fontId = r->getFontID();
		if (r->getStyleFlagsHasXOffset()) 	xoffset = r->getXOffset(); 
		if (r->getStyleFlagsHasYOffset()) 	yoffset = r->getYOffset();
		for (int i=0; i< r->getGlyphCount(); i++) {
			int glyphEntry = r->getGlyphEntries()[i].glyphIndex;
			LocalFontEntry *fontEntry = localFontManager->getChararacterTexture(fontId,glyphEntry);
			if (prevTextureID == -1 || prevTextureID != (int)fontEntry->textureID) {
				if (prevTextureID != -1) {
					drawingEntry->nbElements = iSeg;
					drawingSequence->addDrawingEntry(drawingEntry);
					DrawingEntry *odr = drawingEntry;
					drawingEntry = new DrawingEntry();
					memcpy(drawingEntry,odr,sizeof(DrawingEntry));
				}
				drawingEntry->texNum = fontEntry->textureID;
				drawingEntry->allocateIndexes(6*size);
				indexes = drawingEntry->getIndexes();
				drawingEntry->allocateTexCoords(2*nbPoints);
				texCoords = drawingEntry->getTexCoords();
				iSeg = 0;
				prevTextureID = fontEntry->textureID;
			}
			if (fontEntry != NULL) {
				indexes[iSeg*6+0] = iPoint;
				indexes[iSeg*6+1] = iPoint+1;
				indexes[iSeg*6+2] = iPoint+2;
				indexes[iSeg*6+3] = iPoint+0;
				indexes[iSeg*6+4] = iPoint+2;
				indexes[iSeg*6+5] = iPoint+3;
				iSeg++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[0] + xoffset+dx;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[0] + yoffset-fHeight;
				texCoords[iPoint*2]  = fontEntry->tcx[0] ;
				texCoords[iPoint*2+1]= fontEntry->tcy[0] ;
				iPoint++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[1] + xoffset+dx;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[0] + yoffset-fHeight;
				texCoords[iPoint*2]  = fontEntry->tcx[1] ;
				texCoords[iPoint*2+1]= fontEntry->tcy[0] ;
				iPoint++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[1] + xoffset+dx;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[1] + yoffset-fHeight;
				texCoords[iPoint*2]  = fontEntry->tcx[1] ;
				texCoords[iPoint*2+1]= fontEntry->tcy[1] ;
				iPoint++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[0] + xoffset+dx;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[1] + yoffset-fHeight;
				texCoords[iPoint*2]  = fontEntry->tcx[0] ;
				texCoords[iPoint*2+1]= fontEntry->tcy[1] ;
				iPoint++;
			}
			dx += r->getGlyphEntries()[i].glyphAdvance;
		}
	}
	drawingEntry->nbElements = iSeg;
	drawingSequence->addDrawingEntry(drawingEntry);
	drawingSequence->setNbPoints(nbPoints);
	drawingSequence->setMatrix(tag->getTextMatrix());
	insertSequence(unid,drawingSequence);
}

void Graphics::registerStaticString(long unid, DefineText2Tag* tag) {
	registerStaticString(unid,(DefineTextTag*)tag);
}


/**
 * Register a dynamic Text Area from a TextField object 
 * We pass a ScriptObject because of include constrains, but the parameter is an ASTextField
 **/
void Graphics::registerDynamicString(ScriptObject *os){
	ASTextField  *textField = (ASTextField *)os;
	wstring s = MEMBER_TOSTRING_NOCLONE(textField,L"text");

	int unid = textField->getUnid();
	// dynamic string are rebuilt when user changes the content
	if (drawingSequences[unid] != NULL) clearSequence(unid);
	int 	fontId = int(MEMBER_TONUMERIC_NOCLONE(textField,L"fontId"));
	float 	fHeight = int(MEMBER_TONUMERIC_NOCLONE(textField->getDirectTextFormat(),L"size"));
	int	prevTextureID = -1;
	DrawingSequence	*drawingSequence = new DrawingSequence();
	int nbPoints = 4*s.size();

	drawingSequence->allocateVertices(8+8+2*nbPoints);
	float *fpoints = drawingSequence->getVertices();
	unsigned int *indexes = NULL;
	int iPoint = 0;
	int iSeg = 0;
	DrawingEntry *drawingEntry = new DrawingEntry();
	drawingEntry->allocateTexCoords(8+8+2*nbPoints);
	float *texCoords = drawingEntry->getTexCoords();

	int leftMargin = int(MEMBER_TONUMERIC_NOCLONE(textField->getDirectTextFormat(),L"leftMargin"));
	int rightMargin = int(MEMBER_TONUMERIC_NOCLONE(textField->getDirectTextFormat(),L"rightMargin"));
	int leading = int(MEMBER_TONUMERIC_NOCLONE(textField->getDirectTextFormat(),L"leading"));
	float yoffset=fHeight;
	float xoffset=leftMargin;
	float dx=0;
	int activeWidth = textField->getBounds()->xmax - textField->getBounds()->xmin;
	activeWidth = activeWidth - leftMargin - rightMargin;
	int	xAlign =  MEMBER_TOSTRING_NOCLONE(textField,L"align") == L"true" ? textField->getBounds()->xmin : 0; // TODO: process proper align
	int	yAlign =  MEMBER_TOSTRING_NOCLONE(textField,L"align") == L"true" ? textField->getBounds()->ymin : 0; // TODO: process proper align
	RGBA textColor;
	textColor.red = textColor.green = textColor.blue  = 0; textColor.alpha = 255;
	// Ident???

	BooleanObject *a = (BooleanObject*)textField->getMember(L"border"); 
	bool hasBorder =  a->getValue();
	delete a;
	if (hasBorder) {
		int nbTris = 2;
		drawingEntry->allocateIndexes(3*nbTris);
		indexes = drawingEntry->getIndexes();
		RECT r = *textField->getBounds();
		RGBA c ; c.red= c.green=c.blue=c.alpha=255;
		texture->registerColor(&c);
		int loc = texture->getColorTextureIndex(&c);
		float tx = (loc%256)/256.0f;
		float ty = float(loc/256)/256.0f;
		fpoints[iPoint*2] = r.xmin; fpoints[iPoint*2+1] = r.ymin; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		fpoints[iPoint*2] = r.xmax; fpoints[iPoint*2+1] = r.ymin; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		fpoints[iPoint*2] = r.xmax; fpoints[iPoint*2+1] = r.ymax; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		fpoints[iPoint*2] = r.xmin; fpoints[iPoint*2+1] = r.ymax; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		indexes[0] = iPoint-4;  indexes[1] = iPoint-3; indexes[2] = iPoint-2;
		indexes[3] = iPoint-4;  indexes[4] = iPoint-2; indexes[5] = iPoint-1; 

		drawingEntry->drawingType = DrawingTriangles;
		drawingEntry->fillType = 0;
		drawingEntry->texNum = texture->getColorTexture();
		drawingEntry->nbElements = nbTris;
		drawingSequence->addDrawingEntry(drawingEntry);

		drawingEntry = new DrawingEntry();
		drawingEntry->allocateTexCoords(16+2*nbPoints);
		texCoords = drawingEntry->getTexCoords();
		indexes = NULL;
	}

	for (wstring::size_type i=0; i< s.size(); i++) {
		if (s[i] == '\r' || s[i] == '\n') {
			if (yoffset + fHeight < (textField->getBounds()->ymax - textField->getBounds()->ymin)) {
				yoffset += fHeight;
				dx = 0;
				yoffset += leading;
			}
			continue;
		}

		drawingEntry->drawingType = DrawingFontChars;
		drawingEntry->fillType = 0;

		int color = UI32(MEMBER_TONUMERIC_NOCLONE(textField,L"textColor"));
		textColor.red   = (color>>24)&0xff;
		textColor.green = (color>>16)&0xff;
		textColor.blue  = (color>>8)&0xff;
		textColor.alpha  = int(MEMBER_TONUMERIC_NOCLONE(textField,L"_alpha"));
		drawingEntry->charColor = textColor;

		LocalFontEntry *fontEntry = localFontManager->getChararacterTexture(fontId,s.at(i));
		if (prevTextureID == -1 || (fontEntry != NULL && prevTextureID != (int)fontEntry->textureID)) {
			if (prevTextureID != -1) {
				drawingEntry->nbElements = iSeg;
				drawingSequence->addDrawingEntry(drawingEntry);

				drawingEntry = new DrawingEntry();
				drawingEntry->charColor = textColor;
				drawingEntry->drawingType = DrawingFontChars;
				drawingEntry->fillType = 0;
				if (prevTextureID != -1) {
					drawingEntry->allocateTexCoords(8+8+2*nbPoints);
					texCoords = drawingEntry->getTexCoords();
				}
			}
			if (fontEntry != NULL) {
				drawingEntry->texNum = fontEntry->textureID;
				prevTextureID = fontEntry->textureID;
				drawingEntry->allocateIndexes(6*(s.size()-i));
				indexes = drawingEntry->getIndexes();
			} else {
				if (indexes == NULL) {
					drawingEntry->allocateIndexes(6*(s.size()-i));
					indexes = drawingEntry->getIndexes();
				}
			}
			iSeg = 0;
		}
		if (fontEntry != NULL) {
			float fxincr = fontEntry->advance[0]*fHeight/64.0;
			float l = fHeight*fontEntry->x[1];
			if (dx < activeWidth ) {
				float sub = 0;
				if (dx > activeWidth - l) sub = (dx +l) - activeWidth;
				indexes[iSeg*6+0] = iPoint;
				indexes[iSeg*6+1] = iPoint+1;
				indexes[iSeg*6+2] = iPoint+2;
				indexes[iSeg*6+3] = iPoint+0;
				indexes[iSeg*6+4] = iPoint+2;
				indexes[iSeg*6+5] = iPoint+3;
				iSeg++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[0] + xoffset+dx+xAlign;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[0] + yoffset-fHeight+yAlign;
				texCoords[iPoint*2]  = fontEntry->tcx[0] ;
				texCoords[iPoint*2+1]= fontEntry->tcy[0] ;
				iPoint++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[1] + xoffset+dx-sub+xAlign;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[0] + yoffset-fHeight+yAlign;
				texCoords[iPoint*2]  = fontEntry->tcx[1] - sub/fxincr*(fontEntry->tcx[1]-fontEntry->tcx[0]);
				texCoords[iPoint*2+1]= fontEntry->tcy[0] ;
				iPoint++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[1] + xoffset+dx-sub+xAlign;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[1] + yoffset-fHeight+yAlign;
				texCoords[iPoint*2]  = fontEntry->tcx[1] - sub/fxincr*(fontEntry->tcx[1]-fontEntry->tcx[0]);
				texCoords[iPoint*2+1]= fontEntry->tcy[1] ;
				iPoint++;
				fpoints[iPoint*2]    = fHeight*fontEntry->x[0] + xoffset+dx+xAlign;
				fpoints[iPoint*2+1]  = fHeight*fontEntry->y[1] + yoffset-fHeight+yAlign;
				texCoords[iPoint*2]  = fontEntry->tcx[0] ;
				texCoords[iPoint*2+1]= fontEntry->tcy[1] ;
				iPoint++;
			}
			dx += fxincr;
			// Test breaking line characters
			if (s[i] == ' ' || s[i] == '-') {
				// Check if next word will fit before the end of the line
				float xsize = 0;
				int k = i+1;
				while (k< (int)s.size() && s[k] != ' ' && s[k] != '-') {
					LocalFontEntry *f = localFontManager->getChararacterTexture(fontId,(unsigned char)s[k]);
					if ( f!= NULL) xsize += f->advance[0]*fHeight/64.0;
					k++;
				}
				if (dx+ xsize > activeWidth -l+fxincr) {
					// Next word will not fit in current line
					// If we have one more line, use it
					if (yoffset + fHeight < (textField->getBounds()->ymax - textField->getBounds()->ymin)) {
						yoffset += fHeight;
						dx = 0;
						yoffset += leading;
					}
				}
			}
			if (dx > activeWidth -l+fxincr) {
				if (yoffset + fHeight < (textField->getBounds()->ymax - textField->getBounds()->ymin)) {
					yoffset += fHeight;
					dx = 0;
					yoffset += leading;
				}
			}
		}
	}
	drawingEntry->nbElements = iSeg;
	drawingSequence->addDrawingEntry(drawingEntry);
	drawingSequence->setNbPoints(nbPoints);
	drawingSequence->setMatrix(NULL);

	// Draw EditArea Border
	if (hasBorder ) {
		int nbLines = 4;
		DrawingEntry *drawingEntry = new DrawingEntry();
		drawingEntry->allocateIndexes(2*nbLines);
		indexes = drawingEntry->getIndexes();
		drawingEntry->allocateTexCoords(16+2*nbPoints);
		texCoords = drawingEntry->getTexCoords();
		RECT r = *textField->getBounds();
		texture->registerColor(&textColor);
		int loc = texture->getColorTextureIndex(&textColor);
		float tx = (loc%256)/256.0f;
		float ty = float(loc/256)/256.0f;
		fpoints[iPoint*2] = r.xmin; fpoints[iPoint*2+1] = r.ymin; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		fpoints[iPoint*2] = r.xmax; fpoints[iPoint*2+1] = r.ymin; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		fpoints[iPoint*2] = r.xmax; fpoints[iPoint*2+1] = r.ymax; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		fpoints[iPoint*2] = r.xmin; fpoints[iPoint*2+1] = r.ymax; 
		texCoords[iPoint*2] = tx; texCoords[iPoint*2+1] = ty; 
		iPoint++;
		indexes[0] = iPoint-4;  indexes[1] = iPoint-3; indexes[2] = iPoint-3; indexes[3] = iPoint-2;
		indexes[4] = iPoint-2;  indexes[5] = iPoint-1; indexes[6] = iPoint-1; indexes[7] = iPoint-4;

		drawingEntry->drawingType = DrawingLines;
		drawingEntry->fillType = 0;
		drawingEntry->texNum = texture->getColorTexture();
		drawingEntry->lineWidth = 1.0;
		drawingEntry->nbElements = nbLines;
		drawingSequence->addDrawingEntry(drawingEntry);
	}

	insertSequence(textField->getUnid(),drawingSequence);
}

/**
 * Returns the location of the editing caret
 **/
Caret Graphics::getCaretLocation(DefineEditTextTag* tag, string s, float mouseX, float mouseY) {
	int   fontId = tag->getFontID();
	float fHeight = tag->getFontHeight();
	float yoffset=fHeight;
	float xoffset=tag->getHasLayout() ? tag->getLeftMargin() : 0;
	float dx=0;
	int activeWidth = tag->getBounds().xmax - tag->getBounds().xmin;
	if (tag->getHasLayout())  activeWidth = activeWidth- tag->getLeftMargin() - tag->getRightMargin();
	// Ident???
	Caret res;
	res.height = fHeight;
	res.x = xoffset;
	res.y = yoffset;
	res.charOffset = 0;
	
	for (string::size_type i=0; i< s.size(); i++) {
		if (s[i] == '\r') {
			if (yoffset + fHeight < (tag->getBounds().ymax - tag->getBounds().ymin)) {
				yoffset += fHeight;
				dx = 0;
				if (tag->getHasLayout()) yoffset += tag->getLeading();
			}
			continue;
		}

		LocalFontEntry *fontEntry = localFontManager->getChararacterTexture(fontId,s[i]);
		if (fontEntry != NULL) {
			float adv = fontEntry->advance[0]*fHeight/64;
			if (fabs(mouseY-yoffset+fHeight*0.5) < fHeight/2 && fabs(mouseX-xoffset-dx) < adv) {
				res.x = xoffset+dx;
				res.y = yoffset-fHeight;
				res.charOffset = i;
				return res;
			}
			dx += adv;
			if (tag->getMultiline() && dx > activeWidth) {
				if (yoffset + fHeight < (tag->getBounds().ymax - tag->getBounds().ymin)) {
					yoffset += fHeight;
					dx = 0;
					if (tag->getHasLayout()) yoffset += tag->getLeading();
				}
			}
		}
	}
	return res;
}

/**
 * Construct a morphed shape.
 ***/
void Graphics::buildMorphedShape(int shapeId, DefineMorphShapeTag *tag, float t) {
	int cachedSequenceId = tag->getCharacterID() *10000 + int (t*10000);
	if (getSequence(cachedSequenceId) != NULL) {
		drawingSequences[shapeId] = drawingSequences[cachedSequenceId];
		return;
	}
	//printf("Build MORPH SHAPE :%d\n",tag->getCharacterID());
	// Build rec
	RECT rect;
	rect.xmin = long(tag->getStartBounds()->xmin + t*float(tag->getEndBounds()->xmin - tag->getStartBounds()->xmin));
	rect.xmax = long(tag->getStartBounds()->xmax + t*float(tag->getEndBounds()->xmax - tag->getStartBounds()->xmax));
	rect.ymin = long(tag->getStartBounds()->ymin + t*float(tag->getEndBounds()->ymin - tag->getStartBounds()->ymin));
	rect.ymax = long(tag->getStartBounds()->ymax + t*float(tag->getEndBounds()->ymax - tag->getStartBounds()->ymax));

	// Build Shape
	Shape *s = tag->getStartEdges();
	Shape *e = tag->getEndEdges();

	int prevPoint=0;
	int	newPoint;
	int nbContours=0;
	float x=0,y=0,x1=0,x2=0,y1=0,y2=0;
	StyleChangeRecord *sDef = NULL;
	ShapeWithStyle *swt = new ShapeWithStyle();
	setMorpFillStyle(swt,tag,t);
	setMorpLineStyle(swt,tag,t);
	int originalFillStyleIndex0=-1, originalFillStyleIndex1=-1;
	int colorFillStyleIndex0=-1, colorFillStyleIndex1=-1;
	FILLSTYLE *fill = NULL;
	this->characterId = characterId;
	list<ShapeRecord*>::const_iterator it;
	this->rect = &rect;
	LINESTYLE *lineColor=NULL;
	allowLowPrecision = true;
	allocateShapeProcessObjects(s);
	bool hasTriangles = false;
	bool hasLines = false;
	bool reverseOrder = false;
	int  nb_edges=0;
	list<ShapeRecord*>::const_iterator it2 = e->getShapeRecords()->begin();

	for (it = s->getShapeRecords()->begin(); it != s->getShapeRecords()->end(); it++) {
		UI8 type = (*it)->getTypeFlag();
		switch (type) {
			case StyleChangeRecordType:	{
				StyleChangeRecord *s1 = (StyleChangeRecord*) *it;
				StyleChangeRecord *s2 = NULL;
				if ((*it2)->getTypeFlag() ==  StyleChangeRecordType) s2 = (StyleChangeRecord*) *it2;
				if (s1->getStateNewStyles()) sDef = s1; 

				if (s1->getStateFillStyle0()) {
					int idx = s1->getFillStyle0();
					if (idx > 0) {
						if (sDef != NULL) {
							fill = &sDef->getFillStyles()[idx-1];
						} else {
							if (swt != NULL) fill = &swt->getFillStyles()[idx-1];
						}
						if (fill!=NULL) {
							originalFillStyleIndex0 = idx-1;
							colorFillStyleIndex0 = findInsertColor(fill);
							registerComputedTexture(fill->getFillStyleType(),colorFillStyleIndex0,fill);
						}
						hasTriangles = true;
					} else {
						originalFillStyleIndex0 = -1;
					}			
				}

				if (s1->getStateFillStyle1()) {
					int idx = s1->getFillStyle1();
					if (idx > 0) {
						if (sDef != NULL) {
							 fill = &sDef->getFillStyles()[idx-1];
						} else {
							if (swt != NULL) fill = &swt->getFillStyles()[idx-1];
						}
						if (fill!=NULL) {
							originalFillStyleIndex1 = idx-1;
							colorFillStyleIndex1 = findInsertColor(fill);
							registerComputedTexture(fill->getFillStyleType(),colorFillStyleIndex1,fill);
						}
						hasTriangles = true;
					} else {
						originalFillStyleIndex1 = -1;
					}
				}

				if (s1->getStateLineStyle()) {
					int idx = s1->getLineStyle();
					if (idx > 0) {
						if (sDef != NULL) {
							lineColor = &sDef->getLineStyles()[idx-1];
						} else {
							lineColor = &swt->getLineStyles()[idx-1];
						}
						hasLines = true;
					} else {
						lineColor= NULL;
					}
				} 

				if (s1->getStateMoveTo()) {
					if (nb_edges == 0 && nbSegments > 0 &&nbPoints>0) { // Flush current path
						updateSegments();
						puts("Flush");
						if (hasTriangles) insertSequence(cachedSequenceId,shape2Triangles());
						if (hasLines) insertSequence(cachedSequenceId,shape2Lines());
						prevPoint = newPoint= nbContours = 	nbSegments = nbPoints = 0;
						reverseOrder = false;
						hasLines = false;
					}

					x1 = s1->getMoveDeltaX();
					y1 = s1->getMoveDeltaY();
					if (s2 != NULL) {
						x2 = s2->getMoveDeltaX();
						y2 = s2->getMoveDeltaY();
					}
					x = x1*(1-t) + t*x2;
					y = y1*(1-t) + t*y2;
					prevPoint = findInsertPoint(x,y);
					nbContours++;
					nb_edges = 0;
				}
				if (s1->getStateMoveTo() || (s2!= NULL && s2->getStateMoveTo()) ) 	it2++;
				break; 
			}
			case StraightEdgeRecordType: 
			case CurvedEdgeRecordType: {
				float cx1,cy1,ex1,ey1;
				if ((*it)->getTypeFlag() == CurvedEdgeRecordType) {
					CurvedEdgeRecord *s2 = (CurvedEdgeRecord*) *it;
					cx1 = x1 + s2->getControlDeltaX();
					cy1 = y1 + s2->getControlDeltaY();
					ex1 = cx1 + s2->getAnchorDeltaX();
					ey1 = cy1 + s2->getAnchorDeltaY();
				} else {
					StraightEdgeRecord *s2 = (StraightEdgeRecord*) *it;
					float nx = 0, ny = 0;
					if (s2->getGeneralLineFlag()) {
						nx = s2->getDeltaX();
						ny = s2->getDeltaY();
					} else {
						if (s2->getVertLineFlag()) {
							ny = s2->getDeltaVert();
						} else {
							nx = s2->getDeltaVert();
						}
					}
					cx1 = x1 + nx/2;
					cy1 = y1 + ny/2;
					ex1 = cx1 + nx/2;
					ey1 = cy1 + ny/2;
				}
				float cx2=0,cy2=0,ex2=0,ey2=0;
				if (*it2 != NULL && (*it2)->getTypeFlag() == CurvedEdgeRecordType) {
					CurvedEdgeRecord *s2 = (CurvedEdgeRecord*) *it2;
					cx2 = x2 + s2->getControlDeltaX();
					cy2 = y2 + s2->getControlDeltaY();
					ex2 = cx2 + s2->getAnchorDeltaX();
					ey2 = cy2 + s2->getAnchorDeltaY();
				} else if (*it2 != NULL) {
					StraightEdgeRecord *s2 = (StraightEdgeRecord*) *it2;
					float nx = 0, ny = 0;
					if (s2->getGeneralLineFlag()) {
						nx = s2->getDeltaX();
						ny = s2->getDeltaY();
					} else {
						if (s2->getVertLineFlag()) {
							ny = s2->getDeltaVert();
						} else {
							nx = s2->getDeltaVert();
						}
					}
					cx2 = x2 + nx/2;
					cy2 = y2 + ny/2;
					ex2 = cx2 + nx/2;
					ey2 = cy2 + ny/2;
				}
				
				for (int i=0; i < 5; i++) {
					float tx = (float)i * 0.2f;
					float curX1 = QuadraticBezierInterp(x1,cx1,ex1,tx);
					float curY1 = QuadraticBezierInterp(y1,cy1,ey1,tx);
					float curX2 = QuadraticBezierInterp(x2,cx2,ex2,tx);
					float curY2 = QuadraticBezierInterp(y2,cy2,ey2,tx);
					float curX = curX1*(1-t)+t*curX2;
					float curY = curY1*(1-t)+t*curY2;

					newPoint = findInsertPoint(curX,curY);
					if (prevPoint != newPoint) {
						Segment *s = (Segment*) &segments[nbSegments];
						s->points[0] = prevPoint;
						s->points[1] = newPoint;
						s->originalIndex[0] = originalFillStyleIndex0;
						s->originalIndex[1] = originalFillStyleIndex1;
						s->colorsIndex[0] = colorFillStyleIndex0;
						s->colorsIndex[1] = colorFillStyleIndex1;
						s->lineColor = lineColor;
						nbSegments++;
						prevPoint = newPoint;
					}
				}
				x1 = ex1; x2 = ex2; y1 = ey1; y2=ey2;
				x = x1*(1-t)+t*x2;
				y = y1*(1-t)+t*y2;
				newPoint = findInsertPoint(x,y);
				if (prevPoint != newPoint) {
					Segment *s = (Segment*) &segments[nbSegments];
					s->points[0] = prevPoint;
					s->points[1] = newPoint;
					s->originalIndex[0] = originalFillStyleIndex0;
					s->originalIndex[1] = originalFillStyleIndex1;
					s->colorsIndex[0] = colorFillStyleIndex0;
					s->colorsIndex[1] = colorFillStyleIndex1;
					s->lineColor = lineColor;
					nbSegments++;
					prevPoint = newPoint;
				}
				it2++;
				nb_edges++;
				break; 
			}
		}
	}

	if (nbContours == 0 && nbPoints>0 ) {	// Force closing the figure, so we have a closed area
		Segment *s = (Segment*) &segments[nbSegments];
		s->points[0] = prevPoint;
		s->points[1] = 0;
		s->originalIndex[0] = originalFillStyleIndex0;
		s->originalIndex[1] = originalFillStyleIndex1;
		s->colorsIndex[0] = colorFillStyleIndex0;
		s->colorsIndex[1] = colorFillStyleIndex1;
		s->lineColor = lineColor;
		nbSegments++;
	}
	if (nbSegments > 0 &&nbPoints>0) {
		updateSegments();
		if (hasTriangles) insertSequence(cachedSequenceId,shape2Triangles());
		if (hasLines)	insertSequence(cachedSequenceId,shape2Lines());
	}
	delete []points; points=NULL;
	delete []segments; segments=NULL;
	delete []fillStyles; fillStyles=NULL;
	delete swt;
	drawingSequences[shapeId] = drawingSequences[cachedSequenceId];
}

/**
 * Set Morph fill styles
 **/
void Graphics::setMorpFillStyle(ShapeWithStyle *shape,DefineMorphShapeTag *tag, float t) {
	shape->setNumFillStyles(tag->getMorphFillStyles()->getFillStyleCount());
	FILLSTYLE *f = new FILLSTYLE[shape->getNumFillStyles()];
	for (int i=0; i< shape->getNumFillStyles(); i++) {
		MORPHFILLSTYLE *fs = &tag->getMorphFillStyles()->getFillStyles()[i];
		f[i].setFillStyleType(fs->getFillStyleType());
		int type = f[i].getFillStyleType();
		if (type == 0x0 ) {
			f[i].setColor(interpolateRGBA(fs->getStartColor(),fs->getEndColor(),t));
		}
		if (type == 0x10 || type == 0x12) {
			f[i].setGradientMatrix(interpolateMATRIX(fs->getStartGradientMatrix(),fs->getEndGradientMatrix(),t));
			f[i].setGradient(interpolateMORPHGRADIENT(fs->getGradient(),t));
		}
		if (type == 0x40 || type == 0x41 || type == 0x42 || type == 0x43) {
			f[i].setBitmapId(fs->getBitmapId());
			f[i].setBitmapMatrix(interpolateMATRIX(fs->getStartBitmapMatrix(),fs->getEndBitmapMatrix(),t));
		}		
	}
	shape->setFillStyles(f);
}

/**
 * Set morph line styles
 **/
void Graphics::setMorpLineStyle(ShapeWithStyle *shape,DefineMorphShapeTag *tag, float t) {
	MORPHLINESTYLE *ls = tag->getMorphLineStyles()->getLineStyles();
	shape->setNumLineStyles(tag->getMorphLineStyles()->getLineStyleCount());
	LINESTYLE *l = new LINESTYLE[shape->getNumLineStyles()];
	for (int i=0; i< shape->getNumLineStyles(); i++) {
		l[i].width = int(ls[i].startWidth + t*float(ls[i].endWidth-ls[i].startWidth));
		l[i].color = interpolateRGBA(&ls[i].startColor,&ls[i].endColor,t);
	}
	shape->setLineStyles(l);
}


/**
 * Register a box for drawing video sequences
 **/
void Graphics::registerVideoBox(long unid, int dimx, int dimy, int textureId) {
	DrawingSequence	*drawingSequence = new DrawingSequence();
	drawingSequence->allocateVertices(8);
	float *fpoints = drawingSequence->getVertices();
	DrawingEntry *drawingEntry = new DrawingEntry();
	drawingEntry->allocateIndexes(6);
	unsigned int *indexes = drawingEntry->getIndexes();
	drawingEntry->allocateTexCoords(8);
	float *texCoords = drawingEntry->getTexCoords();
	float a = float(dimx/20)/float(nextPowerOf2(dimx/20));
	float b = float(dimy/20)/float(nextPowerOf2(dimy/20));

	fpoints[0] = 0; fpoints[1] = 0; fpoints[2] = dimx; fpoints[3] = 0; fpoints[4] = dimx; fpoints[5] = dimy; fpoints[6] = 0; fpoints[7] = dimy;
	texCoords[0] = 0; texCoords[1] = 0; texCoords[2] = a; texCoords[3] = 0; texCoords[4] = a; texCoords[5] = b; texCoords[6] = 0; texCoords[7] = b;
	indexes[0] = 0; indexes[1] = 1; indexes[2] = 2; indexes[3] = 0; indexes[4] = 2; indexes[5] = 3;
	drawingEntry->drawingType = DrawingTriangles;
	drawingEntry->texNum = textureId;
	drawingEntry->nbElements = 2;
	drawingEntry->fillType = 0x40;
	drawingSequence->addDrawingEntry(drawingEntry);
	insertSequence(unid,drawingSequence);
}

void Graphics::insertSequence(long characterId, DrawingSequence *d) {
	if (d!=NULL) {
		if (drawingSequences[characterId] == NULL) drawingSequences[characterId] = new ListDrawingSequence();
	 	drawingSequences[characterId]->push_back(d);
	}
}

void Graphics::clearSequence(long unid) {
	for (ListDrawingSequence::const_iterator it=drawingSequences[unid]->begin();it != drawingSequences[unid]->end(); ++it) {
		delete *it;
	}
	drawingSequences[unid]->clear();
	delete drawingSequences[unid];
	drawingSequences.erase(unid);
}

ListDrawingSequence *Graphics::getSequence(int unid) {
	return drawingSequences[unid];
}

void Graphics::setBgColor(RGB c) { 
	RGBA a; a.red=c.red; a.green=c.green; a.blue=c.blue; a.alpha=255;
	defaultColorIndex = texture->registerColor(&a);
	defaultFillStyle.setColor(a);
	defaultFillStyle.setFillStyleType(0);
}

void Graphics::clearSequence(ListDrawingSequence *d) {
	if (d != NULL) {
		for (ListDrawingSequence::const_iterator it2=d->begin();it2 != d->end(); ++it2) 
			delete *it2;
		d->clear();
	}
};

/**
 * duplicate a complete graphic drawing sequence for a given object.
 **/
bool Graphics::cloneDrawingSequence(long unid_src, long unid_dst){
	ListDrawingSequence *src = drawingSequences[unid_src];
	if (src != NULL) {
		ListDrawingSequence *l = new ListDrawingSequence();
		for (ListDrawingSequence::const_iterator i=src->begin(); i != src->end(); ++i) {
			DrawingSequence *ds = *i;
			l->push_back(ds->clone());
		}
		drawingSequences[unid_dst] = l;
	} else {
		//wcout << L"Warning sequence to clone is null!\n";
		return false;
	}
}

DrawingSequence::DrawingSequence() {
	vertices = NULL;
	matrix = NULL;
	clipDepth = 0;
}

DrawingSequence::~DrawingSequence() {
	if (vertices != NULL) {
		free(vertices);
		vertices = NULL;
	}
	if (drawingEntries.size() > 0) {
		for (list<DrawingEntry*>::const_iterator it=drawingEntries.begin();it != drawingEntries.end(); it++) {
			DrawingEntry *d = *it;
			delete d;
		}
	}
	drawingEntries.clear();
}

/**
 * Deep DrawingSequence cloning
 **/
DrawingSequence *DrawingSequence::clone() {
	DrawingSequence *ds = new DrawingSequence(*this);
	ds->clearEntries();
	for (list<DrawingEntry*>::const_iterator it=drawingEntries.begin();it != drawingEntries.end(); ++it) {
		DrawingEntry *d = *it;
		ds->addDrawingEntry(d->clone());
	}
	ds->allocateVertices(this->nbVertices);
	memcpy(ds->getVertices(), getVertices(), this->nbVertices*sizeof(float));
	return ds;
}

DrawingEntry::~DrawingEntry() {
	if (indexes != NULL) { free(indexes); indexes=NULL; }
	if (texCoords != NULL) { free(texCoords); texCoords=NULL; }
}

/**
 * Deep DrawingEntry cloning
 **/
DrawingEntry *DrawingEntry::clone(){
	DrawingEntry *d = new DrawingEntry(*this);
	d->allocateIndexes(nbIndexes);
	memcpy(d->getIndexes(),getIndexes(),nbIndexes*sizeof(unsigned int));
	d->allocateTexCoords(nbTexCoords);
	memcpy(d->getTexCoords(),getTexCoords(),nbTexCoords*sizeof(float));
	return d;
}

void DrawingEntry::allocateIndexes(int nbIndexes) {
	this->nbIndexes = nbIndexes;
	this->indexes = (unsigned int *)malloc(nbIndexes*sizeof(unsigned int));
}

void DrawingEntry::allocateTexCoords(int nbCoords) {
	this->nbTexCoords = nbCoords;
	this->texCoords = (float*)calloc(2*nbCoords*sizeof(float),1);
}

void DrawingSequence::allocateVertices(int nbVertices) {
	this->nbVertices = nbVertices;
	this->vertices = (float*) malloc(nbVertices*sizeof(float));
}
