/**
Macromedia Flash Player - Font management.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "localFonts.hh"
#include "maths.hh"
#include <GL/glx.h>

// Init the library
LocalFontManager::LocalFontManager(BaseTexture *texture) {
	int error = FT_Init_FreeType( &library );
	if ( error ) {
		freeType2Support = false;
	} else {
		freeType2Support = true;
	}
	this->texture = texture;
	defaultFontLoaded= false;
	dpy = NULL;
}

/**
 * The destructor removes the persistant local fonts
 **/
LocalFontManager::~LocalFontManager() {
	for (IntTextureIDListHashMap::iterator it= globalTextureIds.begin(); it != globalTextureIds.end(); it++) {
		TextureIDList *l = (*it).second;
		(*it).second = NULL;
		for (TextureIDList::iterator it2=l->begin(); it2 != l->end(); ++it2) {
			unsigned int a = *it2;
			texture->removeTexture(a);
		}
		//CRASH delete l;
	}
	globalTextureIds.clear();

	for (IntTextureIDListHashMap::iterator it= glyphsTextureIds.begin(); it != glyphsTextureIds.end(); it++) {
		TextureIDList *l = (*it).second;
		(*it).second = NULL;
		for (TextureIDList::iterator it2=l->begin(); it2 != l->end(); ++it2) {
			unsigned int a = *it2;
			texture->removeTexture(a);
		}
		//CRASH delete l;
	}
	glyphsTextureIds.clear();

	if (freeType2Support) FT_Done_FreeType(library);
}

/**
 * Load a subset of chars from TTF and XFonts
 **/
bool LocalFontManager::loadFontFaces(char *fontName,int fontId,bool isBold, bool isItalic, int codeTableSize, UI16* codeTable) {
	if (dpy == NULL) {
		dpy = glXGetCurrentDisplay();
		if (!dpy) return false;
		win = RootWindow(dpy, DefaultScreen(dpy));
	}

	// Try True Type fonts
	bool done = false;
	if (freeType2Support) done=loadTTFaces(fontName,fontId,codeTableSize,codeTable);

	// load and X font
	if (!done) {
		XFontStruct* fs = findXFonts(fontName,isBold,isItalic);
		if (fs != NULL) {
			loadXFontFaces(fs,fontId,codeTableSize,codeTable);
			done = true;
		}
	}
	if (!done) printf("Warning: font %s cannot be found!\n",fontName);
	return done;
}

/**
 * Load full TTF or XFonts
 **/
bool LocalFontManager::loadFullFontFaces(char *fontName,int fontId,bool isBold, bool isItalic) {
	if (dpy == NULL) {
		dpy = glXGetCurrentDisplay();
		if (!dpy) return false;
		win = RootWindow(dpy, DefaultScreen(dpy));
	}

	// Try True Type fonts
	bool done = false;
	if (freeType2Support) done=loadFullTTFaces(fontName,fontId);

	// load and X font
	if (!done) {
		XFontStruct* fs = findXFonts(fontName,isBold,isItalic);
		if (fs != NULL) {
			loadFullXFontFaces(fs,fontId);
			done = true;
		}
	}
	if (!done) printf("Warning: font %s cannot be found!\n",fontName);
	return done;
}

/**
 * Load a set of chars from true type fonts and create textures
 * If it can't load that font, returns false and go back to glyphs rendering
 *
 * Code taken from FTGL.
 **/
bool LocalFontManager::loadTTFaces(char *fontName,int fontId,int codeTableSize, UI16* codeTable) {
	FT_Face face;
	char path[100];
	sprintf(path,"%s/%s.ttf",TTF_PATH,fontName);

	int error = FT_New_Face( library, path, 0, &face );
	if (error) {
		return false;
	}
	error = FT_Set_Char_Size( face, // handle to face object 
               0,		 	// char_width in 1/64th of points 
               16*64, 		// char_height in 1/64th of points 
               300, 		// horizontal device resolution 
               300 ); 		// vertical device resolution 

	int xOffset=0, yOffset=0;
	int textureWidth;
	int textureHeight;
	int padding = 3;
	int glyphWidth=64;
	int glyphHeight=64;
	unsigned int textureID;  
	vector<unsigned int> *textureIDList = new vector<unsigned int>;

	for (int i=0; i< codeTableSize; i++) {
		FT_UInt glyph_index = FT_Get_Char_Index( face, codeTable[i] );
		error = FT_Load_Glyph( face, glyph_index, FT_LOAD_RENDER );
		if (error)  return false;
		FT_GlyphSlot slot = face->glyph;
		FT_Bitmap bitmap  = slot->bitmap;

        if(textureIDList->empty())  {
			CalculateTextureSize(textureWidth,textureHeight,codeTableSize,codeTableSize-i,glyphWidth,glyphHeight,padding);
			textureID = texture->allocateTexture(textureWidth,textureHeight);
            textureIDList->push_back(textureID); 
            xOffset = yOffset = padding;
        }
        
        if( xOffset > ( textureWidth - glyphWidth))  {
            xOffset = padding;
            yOffset += glyphHeight;
            
            if( yOffset > ( textureHeight - glyphHeight)) {
				textureID = texture->allocateTexture(textureWidth,textureHeight);
	            textureIDList->push_back(textureID); 
                yOffset = padding;
            }
        }
        	
		// Store bitmap in texture
		int destWidth  = bitmap.width;
		int	destHeight = bitmap.rows;
    
		if( destWidth && destHeight)	{
			texture->storeAlphaBitmap(textureID,xOffset, yOffset, destWidth, destHeight,bitmap.buffer);
		}

		// Register character
		LocalFontEntry fontEntry;
		fontEntry.textureID = textureID;
		fontEntry.tcx[0] = float(xOffset)/float(textureWidth);
		fontEntry.tcx[1] = float(xOffset+glyphWidth-1)/float(textureWidth);
		fontEntry.tcy[0] = float(yOffset)/float(textureHeight);
		fontEntry.tcy[1] = float(yOffset+destHeight)/float(textureHeight);
		fontEntry.x[0] = float(padding)/float(glyphWidth);
		fontEntry.x[1] = fontEntry.x[0]+float(glyphWidth)/float(glyphWidth);
		fontEntry.y[0] = (-slot->bitmap_top+glyphHeight)/float(glyphHeight);
		fontEntry.y[1] = fontEntry.y[0]+float(bitmap.rows)/float(glyphHeight);
		fontEntry.advance[0] = slot->advance.x>>6; 
		fontEntry.advance[1] = slot->advance.y>>6;
		char name[100];
		sprintf(name,"%d_%d",fontId,i);
		fontEntries[name] = fontEntry;
        xOffset += glyphWidth;
	}
	glyphsTextureIds[fontId] = textureIDList;
	texture->setCharacterDefined(fontId);
	return true;
}

/**
 *
 * Same as above, but for XFonts
 *
 **/
bool LocalFontManager::loadXFontFaces(XFontStruct *fs,int fontId,int codeTableSize,UI16* codeTable) {
	int xOffset=0, yOffset=0;
	int textureWidth;
	int textureHeight;
	int padding = 0;
	int glyphWidth = fs->max_bounds.rbearing - fs->min_bounds.lbearing;
	int glyphHeight = fs->max_bounds.ascent + fs->max_bounds.descent;
	unsigned int textureID;
	vector<unsigned int> *textureIDList = new vector<unsigned int>;
	unsigned char*bitmap = (unsigned char *) malloc((glyphWidth * glyphHeight) * sizeof(unsigned char));
	Pixmap pixmap = XCreatePixmap(dpy, win, 10, 10, 1);
	XGCValues values;
	unsigned long valuemask;
	values.foreground = BlackPixel(dpy, DefaultScreen(dpy));
	values.background = WhitePixel(dpy, DefaultScreen(dpy));
	values.font = fs->fid;
	valuemask = GCForeground | GCBackground | GCFont;
	GC gc = XCreateGC(dpy, pixmap, valuemask, &values);
	XFreePixmap(dpy, pixmap);

	for (int i=0; i< codeTableSize; i++) {
		XCharStruct *ch = isvalid(fs, codeTable[i]);
		if (ch==NULL) continue; // char non supported

        if(textureIDList->empty())  {
			CalculateTextureSize(textureWidth,textureHeight,codeTableSize,codeTableSize-i,glyphWidth,glyphHeight,padding);
			textureID = texture->allocateTexture(textureWidth,textureHeight);
            textureIDList->push_back(textureID); 
            xOffset = yOffset = padding;
        }
        
        if( xOffset > ( textureWidth - glyphWidth))  {
            xOffset = padding;
            yOffset += glyphHeight;
            
            if( yOffset > ( textureHeight - glyphHeight)) {
				textureID = texture->allocateTexture(textureWidth,textureHeight);
	            textureIDList->push_back(textureID); 
                yOffset = padding;
            }
        }
        	
		int x0 = -ch->lbearing;
		int y0 = ch->ascent;
		int destWidth  = glyphWidth;
		int	destHeight = glyphHeight;
		fillXBitmap(gc,glyphWidth,glyphHeight,x0,y0,codeTable[i],bitmap);
		texture->storeAlphaBitmap(textureID,xOffset, yOffset, destWidth, destHeight,bitmap);

		// Register character
		LocalFontEntry fontEntry;
		fontEntry.textureID = textureID;
		fontEntry.tcx[0] = float(xOffset)/float(textureWidth);
		fontEntry.tcx[1] = float(xOffset+glyphWidth-1)/float(textureWidth);
		fontEntry.tcy[0] = float(yOffset)/float(textureHeight);
		fontEntry.tcy[1] = float(yOffset+destHeight)/float(textureHeight);
		fontEntry.x[0] = float(ch->lbearing)/float(glyphWidth);
		fontEntry.x[1] = fontEntry.x[0]+1;
		fontEntry.y[0] = (-ch->ascent+glyphHeight)/float(glyphHeight);
		fontEntry.y[1] = fontEntry.y[0]+1;
		fontEntry.advance[0] = ch->width * 64 / glyphWidth; 
		fontEntry.advance[1] = 0;
		char name[100];
		sprintf(name,"%d_%d",fontId,i);
		fontEntries[name] = fontEntry;
        xOffset += glyphWidth;
	}
	glyphsTextureIds[fontId] = textureIDList;
	texture->setCharacterDefined(fontId);
	free(bitmap);
	return true;
}

/**
 * Load a full font. This is used in conjunction with Dynamic Edit text
 *
 **/
bool LocalFontManager::loadFullTTFaces(char *fontName,int fontId) {
	if (globalTextureIds[fontId] != NULL) return true;	// Already loaded
	FT_Face face;
	char path[200];
	sprintf(path,"%s/%s.ttf",TTF_PATH,fontName);

	int error = FT_New_Face( library, path, 0, &face );
	if ( error) {
		return false;
	}
	error = FT_Set_Char_Size( face, // handle to face object 
               0,		 	// char_width in 1/64th of points 
               16*64, 		// char_height in 1/64th of points 
               300, 		// horizontal device resolution 
               300 ); 		// vertical device resolution 

	int xOffset=0, yOffset=0;
	int textureWidth;
	int textureHeight;
	int padding = 3;
	int glyphWidth=64;
	int glyphHeight=64;
	unsigned int textureID;  
	vector<unsigned int> *textureIDList = new vector<unsigned int>;

	for (int i=0; i< face->num_glyphs; i++) {
		FT_UInt glyph_index = FT_Get_Char_Index( face, i );
		error = FT_Load_Glyph( face, glyph_index, FT_LOAD_RENDER );
		if (error)  return false;
		FT_GlyphSlot slot = face->glyph;
		FT_Bitmap bitmap  = slot->bitmap;

        if( textureIDList->empty())  {
			CalculateTextureSize(textureWidth,textureHeight,face->num_glyphs,face->num_glyphs-i,glyphWidth,glyphHeight,padding);
			textureID = texture->allocateTexture(textureWidth,textureHeight);
            textureIDList->push_back(textureID); 
            xOffset = yOffset = padding;
        }
       
        if( xOffset > ( textureWidth - glyphWidth))  {
            xOffset = padding;
            yOffset += glyphHeight;
            
            if( yOffset > ( textureHeight - glyphHeight)) {
				textureID = texture->allocateTexture(textureWidth,textureHeight);
	            textureIDList->push_back(textureID); 
                yOffset = padding;
            }
        }
	
		// Store bitmap in texture
		int destWidth  = bitmap.width;
		int	destHeight = bitmap.rows;

		if( destWidth && destHeight)	{
			texture->storeAlphaBitmap(textureID,xOffset, yOffset, destWidth, destHeight,bitmap.buffer);
		}


		// Register character
		LocalFontEntry fontEntry;
		fontEntry.textureID = textureID;
		fontEntry.tcx[0] = float(xOffset)/float(textureWidth);
		fontEntry.tcx[1] = float(xOffset+glyphWidth-1)/float(textureWidth);
		fontEntry.tcy[0] = float(yOffset)/float(textureHeight);
		fontEntry.tcy[1] = float(yOffset+destHeight)/float(textureHeight);
		fontEntry.x[0] = float(padding)/float(glyphWidth);
		fontEntry.x[1] = fontEntry.x[0]+float(glyphWidth)/float(glyphWidth);
		fontEntry.y[0] = (-slot->bitmap_top+glyphHeight)/float(glyphHeight);
		fontEntry.y[1] = fontEntry.y[0]+float(bitmap.rows)/float(glyphHeight);
		fontEntry.advance[0] = (slot->advance.x >>6) - 1;
		fontEntry.advance[1] = slot->advance.y >>6;
		char name[200];
		sprintf(name,"%d_%d",fontId,i);
		fontEntries[name] = fontEntry;
		if (!defaultFontLoaded) {
			sprintf(name,"%d_%d",0,i);
			fontEntries[name] = fontEntry;
		}
        xOffset += glyphWidth;
	}

	// For performance reasons, we store local fonts in a seperate set to avoid removal each time the scene is redisplayed
	globalTextureIds[fontId] = textureIDList;
	defaultFontLoaded = true;
	texture->setCharacterDefined(fontId);
	return true;
}

/**
 * Load a complete font from XFonts
 **/
bool LocalFontManager::loadFullXFontFaces(XFontStruct *fs,int fontId) {
	int xOffset=0, yOffset=0;
	int textureWidth;
	int textureHeight;
	int padding = 0;
	int glyphWidth = fs->max_bounds.rbearing - fs->min_bounds.lbearing;
	int glyphHeight = fs->max_bounds.ascent + fs->max_bounds.descent;
	unsigned int textureID;
	vector<unsigned int> *textureIDList = new vector<unsigned int>;
	unsigned char*bitmap = (unsigned char *) malloc((glyphWidth * glyphHeight) * sizeof(unsigned char));
	Pixmap pixmap = XCreatePixmap(dpy, win, 10, 10, 1);
	XGCValues values;
	unsigned long valuemask;
	values.foreground = BlackPixel(dpy, DefaultScreen(dpy));
	values.background = WhitePixel(dpy, DefaultScreen(dpy));
	values.font = fs->fid;
	valuemask = GCForeground | GCBackground | GCFont;
	GC gc = XCreateGC(dpy, pixmap, valuemask, &values);
	XFreePixmap(dpy, pixmap);
	int count=fs->max_char_or_byte2 - fs->min_char_or_byte2;

	for (int i=0; i< count; i++) {
		int c = i+fs->min_char_or_byte2;
		XCharStruct *ch = isvalid(fs, c);
		if (ch == NULL) continue;

        if( textureIDList->empty())  {
			CalculateTextureSize(textureWidth,textureHeight,count,count-i,glyphWidth,glyphHeight,padding);
			textureID = texture->allocateTexture(textureWidth,textureHeight);
            textureIDList->push_back(textureID); 
            xOffset = yOffset = padding;
        }
       
        if( xOffset > ( textureWidth - glyphWidth))  {
            xOffset = padding;
            yOffset += glyphHeight;
            
            if( yOffset > ( textureHeight - glyphHeight)) {
				textureID = texture->allocateTexture(textureWidth,textureHeight);
	            textureIDList->push_back(textureID); 
                yOffset = padding;
            }
        }
        	
		int x0 = -ch->lbearing;
		int y0 = ch->ascent;
		int	destHeight = glyphHeight;
		fillXBitmap(gc,glyphWidth,glyphHeight,x0,y0,c,bitmap);
		texture->storeAlphaBitmap(textureID,xOffset, yOffset, glyphWidth, glyphHeight,bitmap);

		// Register character
		LocalFontEntry fontEntry;
		fontEntry.textureID = textureID;
		fontEntry.tcx[0] = float(xOffset)/float(textureWidth);
		fontEntry.tcx[1] = float(xOffset+glyphWidth-1)/float(textureWidth);
		fontEntry.tcy[0] = float(yOffset)/float(textureHeight);
		fontEntry.tcy[1] = float(yOffset+destHeight-1)/float(textureHeight);
		fontEntry.x[0] = float(ch->lbearing)/float(glyphWidth);
		fontEntry.x[1] = fontEntry.x[0]+1;
		fontEntry.y[0] = (-ch->ascent+glyphHeight)/float(glyphHeight);
		fontEntry.y[1] = fontEntry.y[0]+1;

		fontEntry.advance[0] = ch->width * 64 / glyphWidth; 
		fontEntry.advance[1] = 0;
		char name[200];
		sprintf(name,"%d_%d",fontId,i);
		fontEntries[name] = fontEntry;
		if (!defaultFontLoaded) {
			sprintf(name,"%d_%d",0,i);
			fontEntries[name] = fontEntry;
		}
        xOffset += glyphWidth;
	}

	// For performance reasons, we store local fonts in a seperate set to avoid removal each time the scene is redisplayed
	globalTextureIds[fontId] = textureIDList;
	defaultFontLoaded = true;
	texture->setCharacterDefined(fontId);
	free(bitmap);

	return true;
}

/**
 * Load font from rasterized glyphs (those defined inside a flash file).
 **/
bool LocalFontManager::loadGlyphFaces(int fontId,list<UI8*> *bitmaps) {
	int xOffset=0, yOffset=0;
	int textureWidth;
	int textureHeight;
	int padding = 0;
	int glyphWidth=64;
	int glyphHeight=64;
	list<UI8*>::iterator it = bitmaps->begin();
	int codeTableSize = bitmaps->size();
	unsigned int textureID;
	vector<unsigned int> *textureIDList = new vector<unsigned int>;

	for (int i=0; i< codeTableSize; i++, it++) {
		UI8* bitmap = *it;

        if( textureIDList->empty())  {
			CalculateTextureSize(textureWidth,textureHeight,codeTableSize,codeTableSize-i,glyphWidth,glyphHeight,padding);
			textureID = texture->allocateTexture(textureWidth,textureHeight);
            textureIDList->push_back(textureID); 
            xOffset = yOffset = padding;
        }
        
        if( xOffset > ( textureWidth - glyphWidth))  {
            xOffset = padding;
            yOffset += glyphHeight;
            
            if( yOffset > ( textureHeight - glyphHeight)) {
				textureID = texture->allocateTexture(textureWidth,textureHeight);
                textureIDList->push_back(textureID);
                yOffset = padding;
            }
        }
        	
		// Store bitmap in texture
		int destWidth  = 64;
		int	destHeight = 64;
    
		if( destWidth && destHeight)	{
			texture->storeAlphaBitmap(textureID,xOffset, yOffset, destWidth, destHeight,bitmap);
		}


		// Register character
		LocalFontEntry fontEntry;
		fontEntry.textureID = textureID;
		fontEntry.tcx[0] = float(xOffset)/float(textureWidth);
		fontEntry.tcx[1] = float(xOffset+64)/float(textureWidth);
		fontEntry.tcy[0] = float(yOffset)/float(textureHeight);
		fontEntry.tcy[1] = float(yOffset+64)/float(textureHeight);
		fontEntry.x[0] = -0.125;
		fontEntry.x[1] =  1.000;
		fontEntry.y[0] =  0.125;
		fontEntry.y[1] =  1.25;
		char name[100];
		sprintf(name,"%d_%d",fontId,i);
		fontEntries[name] = fontEntry;
        xOffset += glyphWidth;
	}
	glyphsTextureIds[fontId] = textureIDList;
	texture->setCharacterDefined(fontId);
	return true;
}


void LocalFontManager::CalculateTextureSize(int &textureWidth, int &textureHeight, int numGlyphs,int remGlyphs, int glyphWidth,int glyphHeight, int padding) {
	int maximumGLTextureSize = texture->getMaxTextureSize();   
	textureWidth = nextPowerOf2( (remGlyphs * glyphWidth) + ( padding * 2));
	textureWidth = textureWidth > maximumGLTextureSize ? maximumGLTextureSize : textureWidth;

	int h = static_cast<int>( (textureWidth - ( padding * 2)) / glyphWidth);
        
	textureHeight = nextPowerOf2( (( numGlyphs / h) + 1) * glyphHeight);
	textureHeight = textureHeight > maximumGLTextureSize ? maximumGLTextureSize : textureHeight;
}

/**
 * Load a default font
 **/
void LocalFontManager::loadDefaultFont(){
	if (!defaultFontLoaded) {
		loadFullTTFaces("Verdana.ttf",0);
		defaultFontLoaded = true;
	}
}

/*
 * Get a font entry
 **/
LocalFontEntry *LocalFontManager::getChararacterTexture(int fontId, int glyphsId) {
	char name[100];
	sprintf(name,"%d_%d",fontId,glyphsId);
	if (fontEntries.find(name) == fontEntries.end()) return NULL;
	return &fontEntries[name];
}


/**
 * Try to load the better posible fonts
 **/
XFontStruct* LocalFontManager::findXFonts(string originalName, bool isBold, bool isItalic) {
	XFontStruct* fontInfo = NULL;
	//printf("Request: %s\n",originalName.c_str());
	string src="";
	for (string::size_type i=0; i< originalName.size(); i++) {
		if (originalName[i] >='A' && originalName[i]<='Z') src+=originalName[i]-'A'+'a';
		else src+=originalName[i];
	}

	string mainFamilly = src;
	string::size_type pos = mainFamilly.find('_');
	if (pos != string::npos) mainFamilly = mainFamilly.substr(0,pos);

	// Try to load the better looking font. Goes down from 24 down to 8
	for (int i=34; i>=0; i--) {
		char tmp[4];  sprintf(tmp,"%d",i);
		string fontName = "*"+mainFamilly;
		if (isBold) fontName+="-bold"; else fontName+="-medium";
		if (isItalic) fontName+="-i-"; else fontName+="-r-";
		fontName = fontName+"*--"+tmp+"*";
		//printf("%s\n",fontName.c_str());
		fontInfo = XLoadQueryFont(dpy, fontName.c_str());
		if (fontInfo) return fontInfo;
	}

	// Truncate family extension : Arial Narrow
	pos = mainFamilly.find(' ');
	if (pos != string::npos) mainFamilly = mainFamilly.substr(0,pos);
	for (int i=24; i>=0; i--) {
		char tmp[4];  sprintf(tmp,"%d",i);
		string fontName = "*"+mainFamilly;
		if (isBold) fontName+="-bold"; else fontName+="-medium";
		if (isItalic) fontName+="-i-"; else fontName+="-r-";
		fontName = fontName+"*--"+tmp+"*";
		fontInfo = XLoadQueryFont(dpy, fontName.c_str());
		if (fontInfo) return fontInfo;
	}


	// Same, but remove the bold
	for (int i=24; i>=0; i--) {
		char tmp[4];  sprintf(tmp,"%d",i);
		string fontName = "*"+mainFamilly;
		fontName+="-*";
		if (isItalic) fontName+="-i-"; else fontName+="-r-";
		fontName = fontName+"*--"+tmp+"*";
		fontInfo = XLoadQueryFont(dpy, fontName.c_str());
		if (fontInfo) return fontInfo;
	}

	// Same, but remove the italic
	for (int i=24; i>=0; i--) {
		char tmp[4];  sprintf(tmp,"%d",i);
		string fontName = "*"+mainFamilly;
		fontName = fontName+"*--"+tmp+"*";
		fontInfo = XLoadQueryFont(dpy, fontName.c_str());
		if (fontInfo) return fontInfo;
	}

	// Try something simple
	fontInfo = XLoadQueryFont(dpy, "*");
	if (fontInfo) return fontInfo;
	return NULL;
}

/**
 * Clear Glyphs 
 **/
void LocalFontManager::removeFont(int fontId) {
/**

FIXME. Causing troubles with firefox.

	if (glyphsTextureIds[fontId] != NULL) {
		TextureIDList *l = glyphsTextureIds[fontId];
		for (TextureIDList::iterator it=l->begin(); it != l->end(); it++) {
			unsigned int a = *it;
			glDeleteTextures(1,&a);
		}
		delete l;
		l = NULL;
		glyphsTextureIds.erase(fontId);
	}
*/
}

/*
 * determine if a given glyph is valid and return the corresponding XCharStruct.
 */
XCharStruct *LocalFontManager::isvalid(XFontStruct * fs, unsigned int which) {
	unsigned int rows, pages;
	unsigned int byte1 = 0, byte2 = 0;
	int i, valid = 1;

	rows = fs->max_byte1 - fs->min_byte1 + 1;
	pages = fs->max_char_or_byte2 - fs->min_char_or_byte2 + 1;

	if (rows == 1) { // "linear" fonts 
		if ((fs->min_char_or_byte2 > which) || (fs->max_char_or_byte2 < which))	valid = 0;
	} else { 		// "matrix" fonts 
		byte2 = which & 0xff;
		byte1 = which >> 8;
		if ((fs->min_char_or_byte2 > byte2) || (fs->max_char_or_byte2 < byte2) || (fs->min_byte1 > byte1) || (fs->max_byte1 < byte1)) valid = 0;
	}
	if (valid) {
		if (fs->per_char) {
			if (rows == 1) { // "linear" fonts 
				return (fs->per_char + (which - fs->min_char_or_byte2));
			} else {		//  "matrix" fonts 
				i = ((byte1 - fs->min_byte1) * pages) + (byte2 - fs->min_char_or_byte2);
				return (fs->per_char + i);
			}
		} else {
			return (&fs->min_bounds);
		}
	}
	return (NULL);
}

/*
 * Generate OpenGL-compatible bitmap.
 */
void LocalFontManager::fillXBitmap(GC gc,int width,int height, int x0, int y0,int c, unsigned char *bitmap) {
	Pixmap pixmap = XCreatePixmap(dpy, win, width, height, 1);
	XSetForeground(dpy, gc, 0);
	XFillRectangle(dpy, pixmap, gc, 0, 0, width, height);
	XSetForeground(dpy, gc, 1);

	XChar2b char2b;
	char2b.byte1 = (c >> 8) & 0xff;
	char2b.byte2 = (c & 0xff);
	XDrawString16(dpy, pixmap, gc, x0, y0, &char2b, 1);
	XImage * image = XGetImage(dpy, pixmap, 0, 0, width, height, 1, XYPixmap);
	if (image) {
		for (int y = 0; y < height; y++)
			for (int x = 0; x < width; x++)
				if (XGetPixel(image, x, y)) bitmap[width * y + x] = 255;
				else bitmap[width * y + x] = 0;
		XDestroyImage(image);
	}
   XFreePixmap(dpy, pixmap);
}
