/**
Macromedia Flash OpenGL Texture management implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "ogl_texture.hh"

/**
 * The constructor of the OpenGL implementation of the texture store.
 **/
OglTexture::OglTexture()  : BaseTexture(){
	glColorTexture = 0;

	// Some Nvidia drivers have problem handling 2024x2024. Use that flag to go down to 1024x1024.
	char *ext = (char*)glGetString(GL_VENDOR);
	if (strlen(ext) > 2 && ext[0] == 'N' && ext[1] == 'V' && ext[2] == 'I') useConcervativeTexture = true; else useConcervativeTexture = false;
}

/**
 * Stores a one dim texture to the OpenGL renderer
 **/
unsigned int OglTexture::store1DTextureRepeatLinear(int generatedTexSize, RGBA*textureMemory) {
    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);

    glBindTexture( GL_TEXTURE_1D, textID);
    glEnable(GL_TEXTURE_1D);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage1D( GL_TEXTURE_1D, 0, GL_RGBA, generatedTexSize, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
    glBindTexture(GL_TEXTURE_1D, 0);
    glDisable(GL_TEXTURE_1D);
	return (unsigned int)textID;
}

/**
 * Stores a two dim texture to the OpenGL renderer : REPEAT/LINEAR
 **/
unsigned int OglTexture::store2DTextureRepeatLinear(int width, int height, RGBA*textureMemory) {
    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture(GL_TEXTURE_2D, textID);
    glEnable(GL_TEXTURE_2D);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
	glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
	return (unsigned int)textID;
}

/**
 *  Allocate a texture for local fonts
 **/
unsigned int OglTexture::allocateTexture(int textureWidth, int textureHeight) {   
    int totalMemory = textureWidth * textureHeight;
    unsigned char* textureMemory = new unsigned char[totalMemory];
    memset( textureMemory, 0, totalMemory);

    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture( GL_TEXTURE_2D, textID);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA, textureWidth, textureHeight, 0, GL_ALPHA, GL_UNSIGNED_BYTE, textureMemory);
    glDisable(GL_TEXTURE_2D);

    delete [] textureMemory;
	//printf("Allocate texture %dX%d - 8bits ->id:%d\n",textureWidth, textureHeight,textID);
    return (unsigned int)textID;
}

/**
 *  Allocate a texture for Video Rendering (RGB)
 **/
unsigned int OglTexture::allocateTexture24(int textureWidth, int textureHeight) {   
	//printf("Allocate texture %dX%d - 24bits\n",textureWidth, textureHeight);
    int totalMemory = textureWidth * textureHeight * 3;
    unsigned char* textureMemory = new unsigned char[totalMemory];
    memset( textureMemory, 0xff, totalMemory);

    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture( GL_TEXTURE_2D, textID);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, textureWidth, textureHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, textureMemory);
    glDisable(GL_TEXTURE_2D);

    delete [] textureMemory;
    return (unsigned int)textID;
}

/**
 * To allow full material & ligthing effect (CxFormWithAlpha) colors are  handled as texture coordinates.
 **/
int OglTexture::registerColor(RGBA *c) {
	if (glColorTexture ==0) {
	    int totalMemory = 256*256*4;
	    unsigned char* textureMemory = new unsigned char[totalMemory];
	    memset( textureMemory, 0, totalMemory);	
	    glGenTextures( 1, (GLuint*)&glColorTexture);
	    glBindTexture( GL_TEXTURE_2D, glColorTexture);
	    glEnable(GL_TEXTURE_2D);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 256, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
	    glDisable(GL_TEXTURE_2D);
	    delete [] textureMemory;
	}
	UI32 index = (((((c->red<<8)+c->green)<<8)+c->blue)<<8)+c->alpha;
	if (colorEntries[index] == 0) {
		//printf("%d  (%d,%d,%d,%d) Tid:%d\n",nbColorEntries,c->red,c->green,c->blue,c->alpha,glColorTexture);
		colorEntries[index] = nbColorEntries;
		invColorEntries[nbColorEntries] = index;
	    glEnable(GL_TEXTURE_2D);
		glPushClientAttrib( GL_CLIENT_PIXEL_STORE_BIT);
		glPixelStorei( GL_UNPACK_LSB_FIRST, GL_FALSE);
		glPixelStorei( GL_UNPACK_ROW_LENGTH, 0);
		glPixelStorei( GL_UNPACK_ALIGNMENT, 1);
		glBindTexture( GL_TEXTURE_2D, glColorTexture);
		glTexSubImage2D( GL_TEXTURE_2D, 0, nbColorEntries%256, nbColorEntries/256, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, c);
	    glDisable(GL_TEXTURE_2D);
		glPopClientAttrib();
		nbColorEntries ++;
	}	
	return colorEntries[index];
}


/**
 * Transfer video data to texture
 **/
void OglTexture::video2Texture(int textureId, int dimx, int dimy, char *data) {
    glEnable(GL_TEXTURE_2D);
	glPushClientAttrib( GL_CLIENT_PIXEL_STORE_BIT);
	glPixelStorei( GL_UNPACK_LSB_FIRST, GL_FALSE);
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1);
	glBindTexture( GL_TEXTURE_2D, textureId);
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, dimx, dimy, GL_RGB, GL_UNSIGNED_BYTE, data);
    glDisable(GL_TEXTURE_2D);
}

void OglTexture::removeTexture(unsigned int textureId) {
	GLuint glTextureId = textureId;
	glDeleteTextures(1,&glTextureId); 
}

/**
 * Returns max supported texture size
 **/
int OglTexture::getMaxTextureSize() {
	GLint maximumGLTextureSize;
	glGetIntegerv( GL_MAX_TEXTURE_SIZE, &maximumGLTextureSize);
	if (useConcervativeTexture) maximumGLTextureSize >>=1;
	return maximumGLTextureSize;
}

/**
 * Store a bitmap texture
 **/
void OglTexture::storeAlphaBitmap(unsigned int glTextureID, int xOffset, int yOffset, int destWidth, int destHeight, unsigned char* buffer) {
	glPushClientAttrib( GL_CLIENT_PIXEL_STORE_BIT);
	glPixelStorei( GL_UNPACK_LSB_FIRST, GL_FALSE);
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1);
	glBindTexture( GL_TEXTURE_2D, glTextureID);
	glTexSubImage2D( GL_TEXTURE_2D, 0, xOffset, yOffset, destWidth, destHeight, GL_ALPHA, GL_UNSIGNED_BYTE, buffer);
	glPopClientAttrib();
}

