/**
Timing control for syncing audio and video playback
Called by flashplayer and swf_engine and as_audio.
Hopefullly also from the plugins soon.

Copyright (C) 2005  Dag Hovland (dag.hovland@student.uib.no)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/

#include "sys/time.h"		// For gettimeofday
#include <pthread.h>
#include "common.hh"
#include "swf_timer.hh"

pthread_mutex_t frameNumberLock = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;	// Prevents double rw access to soundFrame
pthread_cond_t	frameNum = PTHREAD_COND_INITIALIZER; 		// Signals frame number changed

pthread_mutex_t streamPlayLock = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;	// Prevents double rw access to soundStreamPlaying
pthread_cond_t	streamPlay = PTHREAD_COND_INITIALIZER; 		// Signals sound stream started or stopped



/**
 * The constructor needs information about frame interval, which is taken from the header of the swf file
 **/
SwfTimer::SwfTimer(int timeBetweenFrames)
	: samples(0),
	freq (0),
	soundStreamSampleCount(0),
	streamLocks(0),
	headerFrameDelay(timeBetweenFrames),			//1000.0/header->frameRate. milliseconds between two frames
	soundSkip(0),
	videoFrame(0)
{
	lockFrame();
	gettimeofday(&lastUpdate, NULL);
	unlockFrame();

	soundStreamPlaying[0] = FINISHED;			// spriteId = 0 => main timeline
	soundFrame[0] = 0;
}

/**
 * The destructor
 **/
SwfTimer::~SwfTimer(){
	stop();	
	pthread_cond_destroy(&frameNum);
	pthread_cond_destroy(&streamPlay);
}

/**
 * Removes all sound information. Called from ASSound::stop() and the destructor
 * This -must- be done on jumps, loops - any interuptions on regular motion through the frames
 */
void SwfTimer::stop(){
	lockStream();
	soundStreamPlaying.clear();
	pthread_cond_broadcast(&streamPlay);
	unlockStream();
	lockFrame();
	pthread_cond_broadcast(&frameNum);
	unlockFrame();
	streamIds.clear();
}

/**
 * Returns number of milliseconds to next video frame should be displayed
 * Called by players to use for waiting for events in XNextEventTimed
 *
*/
int SwfTimer::getTimeOut(){
	if (this->skipVideoFrame()) return 0;

	struct timeval now;
	if (gettimeofday(&now, NULL) != 0){
		fprintf(stderr,"Could not get time!.errno=%i\n",errno);
		return 0;
	}
	lockFrame();
	unsigned int diffsec = static_cast<int>(difftime(lastUpdate.tv_sec, now.tv_sec));
	unsigned int diffusec = now.tv_usec - lastUpdate.tv_usec; 
	unsigned int timeout = headerFrameDelay - (diffsec * 1000 + diffusec/1000);
	lastUpdate = now;
	unlockFrame();
	return timeout;
}

/**
 * Called from video player (flashplayer or plugin)
 * when a new image is shown on screen.
 * This point in time is used by getTimeOut to calculate when next frame should be shown
*/
void SwfTimer::screenUpdate(){
	lockFrame();
	gettimeofday(&lastUpdate, NULL);
	unlockFrame();
}

/**
 * Called by ASSOund::init to set the parameters relevant for timing
*/
void SwfTimer::setSoundParams(int samples, int freq, int soundStreamSampleCount){
	this->samples = samples;
	this->freq = freq;
	this->soundStreamSampleCount = soundStreamSampleCount;
	//headerFrameDelay = soundStreamSampleCount / freq;
}

/**
 * Waits till estimated currently playing audio frame equals last played video frame.
 * This is because this method is called just before the next video frame is displayed
 * Called by players to prevent main timeline audio from lagging behind video.
 */
void SwfTimer::waitForAudio(){
	if(getSoundStreamPlaying(0) != PLAYING) return;
	waitForAudioFrame(0, videoFrame);
	return;
}

/**
 * Waits for audio from main timeline reaching the frame after the specified frame.
 * Called by waitForAudio()
*/
void SwfTimer::waitForAudioFrame(UI16 spriteId, int limitFrame){
	if(getSoundStreamPlaying(0) != PLAYING) return;
	if(getSoundFrame(spriteId) > limitFrame) return;

	lockStream();
	while(soundStreamPlaying[0] != PLAYING){
		pthread_cond_wait(&streamPlay, &streamPlayLock);
	}
	unlockStream();

	// Waiting for setSoundFrame and then checking if next soundFrame is enough	
	struct timeval now;
	struct timespec timeout;
	while(getSoundFrame(spriteId) <= limitFrame && getSoundStreamPlaying(0) != FINISHED){
		if (gettimeofday(&now, NULL) != 0){
			fprintf(stderr,"Could not get time!.errno=%i\n",errno);
			exit(1);
		}
		// Calculates time till stream end in milliseconds and nanoseconds
		// Must be outside lock because of the call to getSoundFrame
		int restTime = headerFrameDelay * (limitFrame - getSoundFrame(spriteId));
		unsigned long int nanosecs = (now.tv_usec + (restTime % 1000)*1000) * 1000;

		timeout.tv_sec = static_cast<int> (now.tv_sec+ restTime / 1000 + nanosecs / 1e9);
		timeout.tv_nsec = static_cast<int> (nanosecs % 1000000000);

		lockFrame();
		pthread_cond_timedwait(&frameNum, &frameNumberLock, &timeout);
		unlockFrame();
		
	}
	return;
}

/**
 * Set to PLAYING by ASSound::processSoundStreamBloc. Set to EXITING by fillAudioHandler, and then to
 * FINISHED by processSoundStreamBloc again.
 * Inidicates status of streaming sound in a sprite. Sprite 0 is main timeline.
*/
void SwfTimer::setSoundStreamPlaying(UI16 spriteId, streamstate play){
	lockStream();
	soundStreamPlaying[spriteId] = play;
	pthread_cond_broadcast(&streamPlay);
	unlockStream();
	lockFrame();
	pthread_cond_broadcast(&frameNum);
	unlockFrame();
}

/**
 * Called to see if there is a soundStream playing in the sprite id
 * sprite 0 is the main timeline
*/
streamstate SwfTimer::getSoundStreamPlaying(UI16 spriteId){
	streamstate playing = FINISHED;
	lockStream();
	if(soundStreamPlaying.find(spriteId) != soundStreamPlaying.end())
		playing = (soundStreamPlaying[spriteId]);
	unlockStream();
	return playing;
}

/**
 * Called from stopStream in ASSound
 * returns when the stream has ended playing (FINISHED). This ensures that the new thread does not start before this.
*/
void SwfTimer::waitForStreamEnd(UI16 spriteId){
	struct timeval now;
	struct timespec timeout = *(new struct timespec);

	lockStream();
	
	while(soundStreamPlaying.find(spriteId) != soundStreamPlaying.end() && soundStreamPlaying[spriteId] == PLAYING){
		if (gettimeofday(&now, NULL) != 0){
			fprintf(stderr,"Could not get time!.errno=%i\n",errno);
			exit(1);
		}
		// Calculates time till stream end in milliseconds and nanoseconds
		int restTime = headerFrameDelay;
		unsigned long int nanosecs = (now.tv_usec + (restTime % 1000)*1000) * 1000;

		timeout.tv_sec = static_cast<int> (now.tv_sec+ restTime / 1000 + nanosecs / 1e9);
		timeout.tv_nsec = static_cast<int> (nanosecs % 1000000000);
		pthread_cond_timedwait(&streamPlay, &streamPlayLock, &timeout);
	}
	unlockStream();
}


/**
 * Set by fillAudioHandler when corresponding frame is copied to SDL audio buffer
 * the varable soudnFrameNumber and frameTime is later used by getSOundFrameNumber to estimate
 * currently played soundStreamBlock
 */
void SwfTimer::setSoundFrame(UI16 spriteId, int num, int skip){
	lockFrame();
	soundFrame[spriteId] = num;
	if(spriteId == 0){
		soundSkip = skip;
		if(gettimeofday(&frameTime, NULL) == -1 || &frameTime == NULL){
			fprintf(stderr,"Could not get time. errno=%i\n", errno);
			exit(1);
		}
	}
	pthread_cond_broadcast(&frameNum);
	unlockFrame();
}

/**
 * Estimating current soundBlockTag being played
 * Uses the time since last frameNumber was set by the playing thread
 * and the information in the soundStreamHeader about average number of blocks per frame
 * and frequency
 *
 * Called from flashplayer to sync audio and video
 */
int SwfTimer::getSoundFrame(UI16 spriteId){ 
	struct timeval  now;

	lockFrame();
	if(gettimeofday(&now, NULL) == -1 || &now == NULL){
		printf("Could not get time. errno=%i\n", errno);
		exit(1);
	}
	int fNum = soundFrame[spriteId];
	// These calculations look strange, but they have to follow this order
	// to prevent overflow. The result represents the number of samples played
	// since  soundFrame was set to its current value
	// Bug: FrameTime is only set for sprite 0 - the main timeline.
	// To be precise it should also be set for the others.
	double sampleInterval = freq * difftime(now.tv_sec, frameTime.tv_sec);
	sampleInterval += ( freq / 100 * (now.tv_usec - frameTime.tv_usec) ) / 10000;

	// Calculating currently played frame
	// samples is the size of the soundcard buffer.
	// soundSkip is the position int the frame of the start of the soundblock actually sent to the card
	// when setSoundFrame was called
	double bufRest = samples - sampleInterval - soundSkip;
	int frameInterval = static_cast<int>(bufRest / soundStreamSampleCount);
	int framePlaying = fNum - frameInterval;

	unlockFrame();
	// If the buffer has become empty, do not update!
	if (bufRest < 0)
		return 0;
	return framePlaying-1;
}

/**
 * Called by players.Retruns true if video is lagging during streamingSound
 * playback and it should skip video frames.
 * 
 * Uses information from sound stream about sound currently played (calls getSoundFrame)
 * and from swf_engine about what video frame is currently treated (call getVideoFrame)
*/
bool SwfTimer::skipVideoFrame(){
	if(getSoundStreamPlaying(0) != PLAYING) return false;
	return getSoundFrame(0) > getVideoFrame()+1;
}

/**
 * Called by SwfEngine::decodeSoundStreamBloc to test if a stream with this id
 * has already been started. This must be done, because timing issues may cause audio
 * stream to finish before the engine / video. In case, the engine would otherwise
 * try to restart a stream which is already played. Thats a bad idea.
 *
 * The ids are chosen incrementally by the decoder. 
 *
 * The ids are only deleted on jumps or loops, as they should never be restarted otherwise
*/
bool SwfTimer::getStreamIdStarted(int streamId){
	return find(streamIds.begin(), streamIds.end(), streamId) != streamIds.end();
}

void SwfTimer::setStreamIdStarted(int streamId){
	streamIds.push_back(streamId);
}

void SwfTimer::lockFrame(){
	int err = pthread_mutex_lock(&frameNumberLock);
	if(err != 0){
		if(err == EINVAL){
			printf("mutex frameNumberLock not properly initialized\n");
			exit(1);
		} else if(err == EDEADLK){
			printf("repeated locking of frameNumberLock\n");
			exit(1);
		} else {		
			printf("Unknown error locking of frameNumberLock\n");
			exit(1);
		}
	}
}


void SwfTimer::unlockFrame(){
	pthread_mutex_unlock(&frameNumberLock);
}

void SwfTimer::lockStream(){
	int err = pthread_mutex_lock(&streamPlayLock);
	if(err != 0){
		if(err == EINVAL){
			printf("mutex streamPlayLock not properly initialized\n");
			exit(1);
		} else if(err == EDEADLK){
			printf("repeated locking of streamPlayLock\n");
			exit(1);
		} else {		
			printf("Unknown error locking of streamPlayLock\n");
			exit(1);
		}
	}
}


void SwfTimer::unlockStream(){
	pthread_mutex_unlock(&streamPlayLock);
}
