/**
Plugin interface. Mozilla / Opera / Konqueror (mozilla plugin mode).
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/

#include <stdio.h>
#include <string.h>
#include "npapi.h"
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include "flash_interface.hh"

#define FLASH_XEVENT_MASK (StructureNotifyMask | ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask)
#define BUFFERSIZE (1024*16)

typedef struct _PluginInstance {
	Display 		*dpy;
	Window 			win;
	Widget 			widget;
	tNSBufferDescriptor nsBuffer;
	int				renderingInterval;
	FlashHandle 	fh;
	unsigned long   moz_xid;	// The XID of the window mozilla has for us
	int				renderingInProgress;
	int				shutDownRequested;
} PluginInstance;

typedef struct {
	PluginInstance*data;
	long 		lastCalledTime;	
	void		*next;
} tTimerEntry;

static tTimerEntry timerList;
static XtIntervalId	timer = 0;
static pthread_mutex_t synchro;
extern int currentX11Cursor;

static void startTimer(PluginInstance* This);
static void flashWakeUp();
static void removeTimerCallBack(PluginInstance* This);
static void addTimerCallBack(PluginInstance* This);


/**
 * Creates a new instance of a plug-in and returns an error value. 
 * NPP_New creates a new instance of your plug-in with MIME type specified by pluginType. The parameter mode is NP_EMBED if the instance was created 
 * by an EMBED tag, or NP_FULL if the instance was created by a separate file. You can allocate any instance-specific private data in instance->pdata 
 * at this time. The NPP pointer is valid until the instance is destroyed. 
 **/
NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16 mode, int16 argc,char* argn[],char* argv[], NPSavedData* saved){
	PluginInstance* This;
	if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;
		
	instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
	This = (PluginInstance*) instance->pdata;

	if (This != NULL) {
		This->fh = flashNew();
		This->dpy = 0;
		This->win = 0;
		This->shutDownRequested = 0;
		This->renderingInProgress = 0;
		This->moz_xid = 0;
		return NPERR_NO_ERROR;
	}
	else return NPERR_OUT_OF_MEMORY_ERROR;
}

/**
 * Request the removal of one flash script.
 **/
NPError NPP_Destroy(NPP instance, NPSavedData** save) {
	PluginInstance* This;
	if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;
	if (This != NULL ) {
		if (This->moz_xid != 0) {
			// Request shutdown. Wait until rendering in progess.
			This->shutDownRequested = 1;		
			while (This->renderingInProgress) usleep(100);
			pthread_mutex_lock(&synchro);

			if (This->fh) {
				closeFlash(This->fh);
				This->fh = 0;
			}
			pthread_mutex_unlock(&synchro);

			NPN_MemFree(instance->pdata);
			instance->pdata = NULL;
			This->moz_xid = 0;
		}
  	}	
	return NPERR_NO_ERROR;
}

/**
 * Start the timer, so we start rendering callbacks
 **/
static void startTimer(PluginInstance* This) {
	if (timer == 0 && This != NULL && This->widget != NULL) {
		XtAppContext ctxt = XtWidgetToApplicationContext(This->widget);
		timer = XtAppAddTimeOut(ctxt, 10, (XtTimerCallbackProc) flashWakeUp, This);
	}
}

/**
 * Add a new timer handler callback
 **/
static void addTimerCallBack(PluginInstance* This) {
	if (timerList.data == NULL) {
		timerList.data = This;
		timerList.next = NULL;
		timerList.lastCalledTime = 0;
	} else {
		tTimerEntry *timerEntry = (tTimerEntry*) malloc(sizeof(tTimerEntry));
		timerEntry->data = This;
		timerEntry->next = NULL; 
		timerEntry->lastCalledTime = 0;
		tTimerEntry *last = &timerList;
		while (last->next != NULL) {
			if (last->data == This) return;	// Already registered
			last=last->next;
		}
		last->next = timerEntry;
	}
	This->renderingInProgress = 0;
	This->shutDownRequested = 0;
}

/**
 * Remove a timer handler callback
 **/
static void removeTimerCallBack(PluginInstance* This) {
	if (timerList.data == This) {
		timerList.data = NULL;
		return;
	}
	tTimerEntry *last = &timerList;
	while (last != NULL) {
		if (last->next != NULL && ((tTimerEntry *)(last->next))->data == This) {
			tTimerEntry *tmp = last->next;
			if (tmp->next == NULL) {
				last->next = NULL;
			}
			free(tmp); 
			return;
		}
		last=last->next;
	}
}

/**
 * Flash event callback
 **/
static void flashEvent(Widget w, XtPointer client_data, XEvent *event, Boolean *dispatch) {
	PluginInstance* This;
	This = (PluginInstance*)client_data;
	int redraw = 0;

	if (This->fh) {
		switch (event->type) { 
        	case KeyPress: {
				KeySym key;
				XLookupString((XKeyEvent *)event, NULL, 0, &key, NULL); 
				char code[5];
				XLookupString((XKeyEvent *)event, code, 5, NULL, NULL); 
				keyboardCallback(This->fh,key,code);
				redraw = 1;
				break;
			}
			case KeyRelease: {
				keyboardReleased(This->fh);
				break;
			}
			case ButtonPress: {
				int x = event->xbutton.x;
				int y = event->xbutton.y;
				redraw = mouseClickCallback(This->fh,x,y,0); // MouseActionPress
				break;
			}
		    case ButtonRelease: {
				int x = event->xbutton.x;
				int y = event->xbutton.y;
				redraw = mouseClickCallback(This->fh,x,y,1); //MouseActionRelease
				break;
			}
			case Expose: {
				//redraw = true;
				break;
			}
		}
	}
}

/**
 * This call back is invoked by the timer callback
 * 
 * With this damm problems of threads this X11, we can't spawn threads to render scenes in parallel
 **/
static void flashWakeUp(PluginInstance* This) {
	struct timeval now;
	gettimeofday(&now,0);
	long nowl = now.tv_sec *1000 + now.tv_usec/1000;
	tTimerEntry *last = &timerList;
	timer = 0;
	while (last != NULL) {
		if (last->data != NULL)  {
			if (last->lastCalledTime == 0 || last->lastCalledTime+last->data->renderingInterval <= nowl) {
				if (last->lastCalledTime == 0) {
					last->lastCalledTime = nowl ; 
				} else {
					last->lastCalledTime+=last->data->renderingInterval;
				}
				//  Looks we can't create threads here.
				if (!last->data->shutDownRequested) {
					last->data->renderingInProgress=1;
					if (last->data->fh != NULL) flashRender(last->data->fh);
					last->data->renderingInProgress=0;
				} else {
					tTimerEntry *tmp = last->next;
					removeTimerCallBack(last->data);
					last->data = NULL;
					last = tmp;
				}
			}
		}
		if (last != NULL) last = last->next;
	}

	Window rootresult,childresult;
	int rootx,rooty,winx,winy;
	unsigned int mask;
	static int lastx = 0x7FFFFFFF,lasty = 0x7FFFFFFF;

	if (XQueryPointer(This->dpy,This->win,&rootresult,&childresult,&rootx,&rooty,&winx,&winy,&mask) && (winx != lastx || winy != lasty))
		mouseClickCallback(This->fh,winx,winy,2);

	lastx = winx; lasty = winy;

	// Restart the timer
	last = &timerList;
	while (last != NULL) {
		if (last->data != NULL) {
			startTimer(last->data);
			return;
		}
		last = last->next;
	}
	return;
}

/**
 * Sets the window in which a plug-in draws, and returns an error value. 
 * NPP_SetWindow informs the plug-in instance specified by instance of the the window denoted by window in which the instance draws. 
 * This NPWindow pointer is valid for the life of the instance, or until NPP_SetWindow is called again with a different value. 
 * Subsequent calls to NPP_SetWindow for a given instance typically indicate that the window has been resized. 
 * If either window or window->window are NULL, the plug-in must not perform any additional graphics operations on the window and should free any
 * resources associated with the window. 
 * On Windows and UNIX, the platform-specific window information specified in window->window is a handle to a subwindow of the Netscape window hierarchy.
 * On the Macintosh, this field points to a NP_Port structure.
 **/
NPError NPP_SetWindow(NPP instance, NPWindow* window) {
	PluginInstance* This;
	NPSetWindowCallbackStruct *ws;

	if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;
	if (window == NULL)	return NPERR_NO_ERROR;

	This = (PluginInstance*) instance->pdata;

	if (This->moz_xid){
		if(This->moz_xid == (unsigned long) window->window) { 	// The window exists already 
		} else {	
			This->moz_xid = (unsigned long) window->window;		// The parent window changed. Reparent accordingly.
		}
	} else {
		This->moz_xid = (unsigned long) window->window;
		ws = (NPSetWindowCallbackStruct *)window->ws_info;
		This->win = (Window) window->window;
		This->dpy = ws->display;
		setWindow(This->fh,This->win,This->dpy, ws->visual);
		This->widget = XtWindowToWidget(This->dpy,This->win);	
		XtAddEventHandler(This->widget, FLASH_XEVENT_MASK, True, (XtEventHandler) flashEvent, (XtPointer)This);
	}
	return NPERR_NO_ERROR;
}

/**
 * Notifies an instance of a new data stream and returns an error value. 
 * NPP_NewStream notifies the instance denoted by instance of the creation of a new stream specifed by stream. 
 * The NPStream* pointer is valid until the stream is destroyed. The MIME type of the stream is provided by the parameter type. 
 * The stream may have been started by Netscape from the file specified in the SRC attribute of the EMBED tag (for an embedded instance) or the file
 * itself (for a full-page instance). A plug-in can also request a stream with function NPN_GetURL. 
 * NPP_DestroyStream is called when the stream completes (either successfully or abnormally). 
 * The plug-in can terminate the stream itself by calling NPN_DestroyStream.
 **/
NPError NPP_NewStream(NPP instance,NPMIMEType type,NPStream *stream,NPBool seekable,uint16 *stype) {
	PluginInstance* This;
	if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;
	return NPERR_NO_ERROR;
}


/**
 * Returns the maximum number of bytes that an instance is prepared to accept from the stream. 
 **/
int32  NPP_WriteReady(NPP instance, NPStream *stream) {
	if (instance != NULL) {
		return BUFFERSIZE;
	}
	return 0;
}

/**
 * Delivers data from a stream and returns the number of bytes written. 
 * NPP_Write is called after a call to NPP_NewStream in which the plug-in requested a normal-mode stream, in which the data in the stream is delivered
 * progressively over a series of calls to NPP_WriteReady and NPP_Write. The function delivers a buffer buf of len bytes of data from the stream
 * identified by stream to the instance. The parameter offset is the logical position of buf from the beginning of the data in the stream. 
 * The function returns the number of bytes written (consumed by the instance). A negative return value causes an error on the stream, which will
 * subsequently be destroyed via a call to NPP_DestroyStream. 
 * Note that a plug-in must consume at least as many bytes as it indicated in the preceeding NPP_WriteReady call. All data consumed must be either
 * processed immediately or copied to memory allocated by the plug-in: the buf parameter is not persistent. 
 **/
int32 NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer) {
	PluginInstance* This;
	if (instance != NULL) {
		This = (PluginInstance*) instance->pdata;
		This->nsBuffer.data = (char*)malloc(len);
		memcpy((void*)This->nsBuffer.data,(char*)buffer,len);
		This->nsBuffer.len = len;
		This->nsBuffer.offset = offset;
		This->nsBuffer.flashHandle = This->fh;
		//if (This->decodeThread != 0) {
		//	int *ret;
		//	pthread_join(This->decodeThread, (void**)ret);
		//}
		//pthread_create(&This->decodeThread,NULL,addBytesToDecoder,(void*)(&(This->nsBuffer)));
		int ret = addBytesToDecoder(&This->nsBuffer);
		if (ret != 0) {
			This->renderingInterval = getTimerInterval(This->fh);
			addTimerCallBack(This);
			startTimer(This);
		}
	}
	return len;
}


/**
 * Indicates the closure and deletion of a stream, and returns an error value. 
 * The NPP_DestroyStream function is called when the stream identified by stream for the plug-in instance denoted by instance will be destroyed. 
 * You should delete any private data allocated in stream->pdata at this time. 
 * The reason the stream was destroyed is indicated by the by the parameter reason. The most common reason code is NPRES_DONE, indicating simply that 
 * the stream completed normally because all data was sent to the instance. 
 * Other possible reason codes are NPRES_USER_BREAK, indicating that the user canceled the stream by clicking the "Stop" button, and NPRES_NETWORK_ERR,
 * indicating that the stream failed due to network problems. The complete list of reason codes is found in npapi.h. 
 **/
NPError NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason) {
	PluginInstance* This;
	if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;
	if (reason != NPERR_NO_ERROR) return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;
	if (This == 0) return NPERR_INVALID_INSTANCE_ERROR;
	return NPERR_NO_ERROR;
}



/**
 *
 **/
char* NPP_GetMIMEDescription(void) {
	return("application/x-shockwave-flash:swf:Flash Plugin;application/x-futuresplash:spl:Future Splash");
}

/**
 *
 **/
NPError NPP_GetValue(void *future, NPPVariable variable, void *value) {
	NPError err = NPERR_NO_ERROR;

	switch (variable) {
		case NPPVpluginNameString:
			*((char **)value) = PLUGIN_NAME;
			break;
		case NPPVpluginDescriptionString:
			*((char **)value) = "Flash Movie player " FLASH_VERSION_STRING
		    " compatible with Shockwave Flash 7.0 "
		    "<P>Shockwave is a trademark of <A HREF=\"http://www.macromedia.com\">Macromedia&reg;</A>"
			"<P>GPLFLash homepage : <A HREF=\"http://gplflash.sf.net\">gplflash.sf.net</A>";
			break;
		default:
			err = NPERR_GENERIC_ERROR;
	}
	return err;
}

/**
 * 
 **/
NPError NPP_Initialize(void) {
    freopen("/dev/tty","w",stdout);
    freopen("/dev/tty","w",stderr);
	pthread_mutex_init(&synchro,NULL);
	currentX11Cursor = XC_left_ptr;
	timerList.data = NULL;
	timerList.lastCalledTime = 0;
	timerList.next = NULL;
    XtToolkitInitialize();
    return NPERR_NO_ERROR;
}

jref NPP_GetJavaClass(){
    return NULL;
}

/**
 * On plugin removal, stop the timer
 **/
void NPP_Shutdown(void) {
	pthread_mutex_lock(&synchro);
	pthread_mutex_unlock(&synchro);
	if (timer) XtRemoveTimeOut(timer);
	timer = 0;
	timerList.data = NULL;
	timerList.lastCalledTime = 0;
	timerList.next = NULL;
}



/*
//static void getUrl( char * url, char *target, void *client_data);//static void flashEvent(Widget w, XtPointer client_data, XEvent *event, Boolean *dispatch);
//static void getSwf(char *url, int level, void *client_data);
//static long parseAttributes(int16 n, char *argn[], char *argv[]);

static void
getUrl( char * url, char *target, void *client_data)
{
	NPP instance;

	instance = (NPP)client_data;
	NPN_GetURL(instance, url, target );
}

static void
getSwf(char *url, int level, void *client_data)
{
	PluginInstance* This;
	NPP instance;

	instance = (NPP)client_data;
	if (instance == 0) return;

	This = (PluginInstance*) instance->pdata;
	if (This == 0) return;
	NPN_GetURL(instance, url, 0);
} 

void 
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
	PluginInstance* This;
	if (instance != NULL)
		This = (PluginInstance*) instance->pdata;
}


void 
NPP_Print(NPP instance, NPPrint* printInfo)
{
	if(printInfo == NULL)
		return;

	if (instance != NULL) {
	
		if (printInfo->mode == NP_FULL) {
			printInfo->print.fullPrint.pluginPrinted = FALSE;
		}
		else {	// If not fullscreen, we must be embedded 
		}
	}
}

static long parseAttributes(int16 n, char *argn[], char *argv[]) {
	int16 i;
	long attributes;

	attributes = 0;
	for(i=0; i<n; i++)	{
		if (!strcasecmp(argn[i],"loop")) {
			if (!strcasecmp(argv[i],"true")) {
				attributes |= PLAYER_LOOP;
			}
		}
		if (!strcasecmp(argn[i],"menu")) {
			if (!strcasecmp(argv[i],"true")) {
				attributes |= PLAYER_MENU;
			}
		}
		if (!strcasecmp(argn[i],"quality")) {
			if (!strcasecmp(argv[i],"high")
			 || !strcasecmp(argv[i],"autohigh")) {
				attributes |= PLAYER_QUALITY;
			}
		}
	}
	return attributes;
}
*/

